""" ProxyDB class is a front-end to the ProxyDB MySQL database.

    Database contains the following tables:

    * ProxyDB_Requests -- a delegation requests storage table for a given proxy Chain
    * ProxyDB_CleanProxies -- table for storing proxies in "clean" form, ie without
      the presence of DIRAC and VOMS extensions.
    * ProxyDB_Proxies -- obsolete table for storing proxies with already added DIRAC
      group extension, it is present only for backward compatibility and is used only
      if ProxyDB_CleanProxies does not have the required proxy.
    * ProxyDB_VOMSProxies -- proxy storage table with VOMS extension already added.
    * ProxyDB_Log -- table with logs.
    * ProxyDB_Tokens -- token storage table for proxy requests.
    * ProxyDB_ExpNotifs -- a table for accumulating proxy expiration notifications.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import time
import random
import hashlib
from six.moves.urllib.request import urlopen

from DIRAC import gConfig, gLogger, S_OK, S_ERROR
from DIRAC.Core.Base.DB import DB
from DIRAC.Core.Utilities import DErrno
from DIRAC.Core.Security import Properties
from DIRAC.Core.Security.VOMS import VOMS
from DIRAC.Core.Security.MyProxy import MyProxy
from DIRAC.Core.Security.X509Request import X509Request  # pylint: disable=import-error
from DIRAC.Core.Security.X509Chain import X509Chain, isPUSPdn  # pylint: disable=import-error
from DIRAC.ConfigurationSystem.Client.Helpers import Registry
from DIRAC.ConfigurationSystem.Client.PathFinder import getDatabaseSection
from DIRAC.FrameworkSystem.Client.NotificationClient import NotificationClient
from DIRAC.Resources.ProxyProvider.ProxyProviderFactory import ProxyProviderFactory

DEFAULT_MAIL_FROM = "proxymanager@diracgrid.org"


class ProxyDB(DB):

    NOTIFICATION_TIMES = [2592000, 1296000]

    def __init__(self, useMyProxy=False, mailFrom=None):
        """
        :param bool useMyProxy: use MyProxy...
        :param str mailFrom: address to use as sender for the expiration reminder emails
        """
        DB.__init__(self, "ProxyDB", "Framework/ProxyDB")
        self._mailFrom = mailFrom if mailFrom else DEFAULT_MAIL_FROM
        self.__defaultRequestLifetime = 300  # 5min
        self.__defaultTokenLifetime = 86400 * 7  # 1 week
        self.__defaultTokenMaxUses = 50
        self.__useMyProxy = useMyProxy
        self._minSecsToAllowStore = 3600
        self.__notifClient = NotificationClient()
        retVal = self.__initializeDB()
        if not retVal["OK"]:
            raise Exception("Can't create tables: %s" % retVal["Message"])
        self.purgeExpiredProxies(sendNotifications=False)
        self.__checkDBVersion()

    def getMyProxyServer(self):
        """Get MyProxy server from configuration

        :return: str
        """
        return gConfig.getValue("/DIRAC/VOPolicy/MyProxyServer", "myproxy.cern.ch")

    def getMyProxyMaxLifeTime(self):
        """Get a maximum of the proxy lifetime delegated by MyProxy

        :return: int -- time in a seconds
        """
        return gConfig.getValue("/DIRAC/VOPolicy/MyProxyMaxDelegationTime", 168) * 3600

    def __initializeDB(self):
        """Create the tables

        :result: S_OK()/S_ERROR()
        """
        retVal = self._query("show tables")
        if not retVal["OK"]:
            return retVal

        tablesInDB = [t[0] for t in retVal["Value"]]
        tablesD = {}

        if "ProxyDB_Requests" not in tablesInDB:
            tablesD["ProxyDB_Requests"] = {
                "Fields": {
                    "Id": "INTEGER AUTO_INCREMENT NOT NULL",
                    "UserDN": "VARCHAR(255) NOT NULL",
                    "Pem": "BLOB",
                    "ExpirationTime": "DATETIME",
                },
                "PrimaryKey": "Id",
            }

        if "ProxyDB_CleanProxies" not in tablesInDB:
            tablesD["ProxyDB_CleanProxies"] = {
                "Fields": {
                    "UserName": "VARCHAR(64) NOT NULL",
                    "UserDN": "VARCHAR(255) NOT NULL",
                    "ProxyProvider": 'VARCHAR(64) DEFAULT "Certificate"',
                    "Pem": "BLOB",
                    "ExpirationTime": "DATETIME",
                },
                "PrimaryKey": ["UserDN", "ProxyProvider"],
            }
        # WARN: Now proxies upload only in ProxyDB_CleanProxies, so this table will not be needed in some future
        if "ProxyDB_Proxies" not in tablesInDB:
            tablesD["ProxyDB_Proxies"] = {
                "Fields": {
                    "UserName": "VARCHAR(64) NOT NULL",
                    "UserDN": "VARCHAR(255) NOT NULL",
                    "UserGroup": "VARCHAR(255) NOT NULL",
                    "Pem": "BLOB",
                    "ExpirationTime": "DATETIME",
                    "PersistentFlag": 'ENUM ("True","False") NOT NULL DEFAULT "True"',
                },
                "PrimaryKey": ["UserDN", "UserGroup"],
            }

        if "ProxyDB_VOMSProxies" not in tablesInDB:
            tablesD["ProxyDB_VOMSProxies"] = {
                "Fields": {
                    "UserName": "VARCHAR(64) NOT NULL",
                    "UserDN": "VARCHAR(255) NOT NULL",
                    "UserGroup": "VARCHAR(255) NOT NULL",
                    "VOMSAttr": "VARCHAR(255) NOT NULL",
                    "Pem": "BLOB",
                    "ExpirationTime": "DATETIME",
                },
                "PrimaryKey": ["UserDN", "UserGroup", "vomsAttr"],
            }

        if "ProxyDB_Log" not in tablesInDB:
            tablesD["ProxyDB_Log"] = {
                "Fields": {
                    "ID": "BIGINT NOT NULL AUTO_INCREMENT",
                    "IssuerDN": "VARCHAR(255) NOT NULL",
                    "IssuerGroup": "VARCHAR(255) NOT NULL",
                    "TargetDN": "VARCHAR(255) NOT NULL",
                    "TargetGroup": "VARCHAR(255) NOT NULL",
                    "Action": "VARCHAR(128) NOT NULL",
                    "Timestamp": "DATETIME",
                },
                "PrimaryKey": "ID",
                "Indexes": {"Timestamp": ["Timestamp"]},
            }

        if "ProxyDB_Tokens" not in tablesInDB:
            tablesD["ProxyDB_Tokens"] = {
                "Fields": {
                    "Token": "VARCHAR(64) NOT NULL",
                    "RequesterDN": "VARCHAR(255) NOT NULL",
                    "RequesterGroup": "VARCHAR(255) NOT NULL",
                    "ExpirationTime": "DATETIME NOT NULL",
                    "UsesLeft": "SMALLINT UNSIGNED DEFAULT 1",
                },
                "PrimaryKey": "Token",
            }

        if "ProxyDB_ExpNotifs" not in tablesInDB:
            tablesD["ProxyDB_ExpNotifs"] = {
                "Fields": {
                    "UserDN": "VARCHAR(255) NOT NULL",
                    "UserGroup": "VARCHAR(255) NOT NULL",
                    "LifeLimit": "INTEGER UNSIGNED DEFAULT 0",
                    "ExpirationTime": "DATETIME NOT NULL",
                },
                "PrimaryKey": ["UserDN", "UserGroup"],
            }

        return self._createTables(tablesD)

    def __addUserNameToTable(self, tableName):
        """Add user name to the table

        :param str tableName: table name

        :return: S_OK()/S_ERROR()
        """
        result = self._update("ALTER TABLE `%s` ADD COLUMN UserName VARCHAR(64) NOT NULL" % tableName)
        if not result["OK"]:
            return result
        result = self._query("SELECT DISTINCT UserName, UserDN FROM `%s`" % tableName)
        if not result["OK"]:
            return result
        data = result["Value"]
        for userName, userDN in data:
            if not userName:
                result = Registry.getUsernameForDN(userDN)
                if not result["OK"]:
                    self.log.error("Could not retrieve username for DN", userDN)
                    continue
                userName = result["Value"]
                try:
                    userName = self._escapeString(userName)["Value"]
                    userDN = self._escapeString(userDN)["Value"]
                except KeyError:
                    self.log.error("Could not escape username or DN", "%s %s" % (userName, userDN))
                    continue
                userName = result["Value"]
                result = self._update("UPDATE `%s` SET UserName=%s WHERE UserDN=%s" % (tableName, userName, userDN))
                if not result["OK"]:
                    self.log.error("Could update username for DN", "%s: %s" % (userDN, result["Message"]))
                    continue
                self.log.info("UserDN %s has user %s" % (userDN, userName))
        return S_OK()

    def __checkDBVersion(self):
        """Check DB tables for empty UserName option

        :return: S_OK()/S_ERROR()
        """
        for tableName in ("ProxyDB_CleanProxies", "ProxyDB_Proxies", "ProxyDB_VOMSProxies"):
            result = self._query("describe `%s`" % tableName)
            if not result["OK"]:
                return result
            if "UserName" not in [row[0] for row in result["Value"]]:
                self.log.notice("Username missing in table %s schema. Adding it" % tableName)
                result = self.__addUserNameToTable(tableName)
                if not result["OK"]:
                    return result

    def generateDelegationRequest(self, proxyChain, userDN):
        """Generate a request and store it for a given proxy Chain

        :param X509Chain() proxyChain: proxy as chain
        :param str userDN: user DN

        :return: S_OK(dict)/S_ERROR() -- dict contain id and proxy as string of the request
        """
        retVal = self._getConnection()
        if not retVal["OK"]:
            return retVal
        connObj = retVal["Value"]
        retVal = proxyChain.generateProxyRequest()
        if not retVal["OK"]:
            return retVal
        request = retVal["Value"]
        retVal = request.dumpRequest()
        if not retVal["OK"]:
            return retVal
        reqStr = retVal["Value"]
        retVal = request.dumpPKey()
        if not retVal["OK"]:
            return retVal
        allStr = reqStr + retVal["Value"]
        try:
            sUserDN = self._escapeString(userDN)["Value"]
            sAllStr = self._escapeString(allStr)["Value"]
        except KeyError:
            return S_ERROR("Cannot escape DN")
        cmd = "INSERT INTO `ProxyDB_Requests` ( Id, UserDN, Pem, ExpirationTime )"
        cmd += " VALUES ( 0, %s, %s, TIMESTAMPADD( SECOND, %d, UTC_TIMESTAMP() ) )" % (
            sUserDN,
            sAllStr,
            int(self.__defaultRequestLifetime),
        )
        retVal = self._update(cmd, conn=connObj)
        if not retVal["OK"]:
            return retVal
        # 99% of the times we will stop here
        if "lastRowId" in retVal:
            return S_OK({"id": retVal["lastRowId"], "request": reqStr})
        # If the lastRowId hack does not work. Get it by hand
        retVal = self._query("SELECT Id FROM `ProxyDB_Requests` WHERE Pem='%s'" % reqStr)
        if not retVal["OK"]:
            return retVal
        data = retVal["Value"]
        if not data:
            return S_ERROR("Insertion of the request in the db didn't work as expected")
        userGroup = proxyChain.getDIRACGroup().get("Value") or "unset"
        self.logAction("request upload", userDN, userGroup, userDN, "any")
        # Here we go!
        return S_OK({"id": data[0][0], "request": reqStr})

    def __retrieveDelegationRequest(self, requestId, userDN):
        """Retrieve a request from the DB

        :param int requestId: id of the request
        :param str userDN: user DN

        :return: S_OK(str)/S_ERROR()
        """
        try:
            sUserDN = self._escapeString(userDN)["Value"]
        except KeyError:
            return S_ERROR("Cannot escape DN")
        cmd = "SELECT Pem FROM `ProxyDB_Requests` WHERE Id = %s AND UserDN = %s" % (requestId, sUserDN)
        retVal = self._query(cmd)
        if not retVal["OK"]:
            return retVal
        data = retVal["Value"]
        if len(data) == 0:
            return S_ERROR("No requests with id %s" % requestId)
        request = X509Request()
        retVal = request.loadAllFromString(data[0][0])
        if not retVal["OK"]:
            return retVal
        return S_OK(request)

    def purgeExpiredRequests(self):
        """Purge expired requests from the db

        :return: S_OK()/S_ERROR()
        """
        cmd = "DELETE FROM `ProxyDB_Requests` WHERE ExpirationTime < UTC_TIMESTAMP()"
        return self._update(cmd)

    def deleteRequest(self, requestId):
        """Delete a request from the db

        :param int requestId: id of the request

        :return: S_OK()/S_ERROR()
        """
        cmd = "DELETE FROM `ProxyDB_Requests` WHERE Id=%s" % requestId
        return self._update(cmd)

    def completeDelegation(self, requestId, userDN, delegatedPem):
        """Complete a delegation and store it in the db

        :param int requestId: id of the request
        :param str userDN: user DN
        :param str delegatedPem: delegated proxy as string

        :return: S_OK()/S_ERROR()
        """
        retVal = self.__retrieveDelegationRequest(requestId, userDN)
        if not retVal["OK"]:
            return retVal
        request = retVal["Value"]
        chain = X509Chain(keyObj=request.getPKey())
        retVal = chain.loadChainFromString(delegatedPem)
        if not retVal["OK"]:
            return retVal
        retVal = chain.isValidProxy()
        if not retVal["OK"]:
            return retVal

        result = chain.isVOMS()
        if result["OK"] and result.get("Value"):
            return S_ERROR("Proxies with VOMS extensions are not allowed to be uploaded")

        retVal = chain.getDIRACGroup(ignoreDefault=True)
        if not retVal["OK"]:
            return retVal
        if retVal["Value"]:
            return S_ERROR("Proxies with DIRAC group extensions not allowed to be uploaded")
        retVal = self.__storeProxy(userDN, chain)
        return self.deleteRequest(requestId) if retVal["OK"] else retVal

    def __storeProxy(self, userDN, chain, proxyProvider=None):
        """Store user proxy into the Proxy repository for a user specified by his
        DN and group or proxy provider.

        :param str userDN: user DN from proxy
        :param X509Chain() chain: proxy chain
        :param str proxyProvider: proxy provider name

        :return: S_OK()/S_ERROR()
        """
        retVal = Registry.getUsernameForDN(userDN)
        if not retVal["OK"]:
            return retVal
        userName = retVal["Value"]

        if not proxyProvider:
            result = Registry.getProxyProvidersForDN(userDN)
            if not result["OK"]:
                return result
            proxyProvider = result.get("Value") and result["Value"][0] or "Certificate"

        # Get remaining secs
        retVal = chain.getRemainingSecs()
        if not retVal["OK"]:
            return retVal
        remainingSecs = retVal["Value"]
        if remainingSecs < self._minSecsToAllowStore:
            return S_ERROR(
                "Cannot store proxy, remaining secs %s is less than %s" % (remainingSecs, self._minSecsToAllowStore)
            )

        # Compare the DNs
        retVal = chain.getIssuerCert()
        if not retVal["OK"]:
            return retVal
        proxyIdentityDN = retVal["Value"].getSubjectDN()["Value"]
        if userDN != proxyIdentityDN:
            msg = "Mismatch in the user DN"
            vMsg = "Proxy says %s and credentials are %s" % (proxyIdentityDN, userDN)
            self.log.error(msg, vMsg)
            return S_ERROR("%s. %s" % (msg, vMsg))

        # Check if its limited
        if chain.isLimitedProxy()["Value"]:
            return S_ERROR("Limited proxies are not allowed to be stored")
        dLeft = int(remainingSecs / 86400)
        hLeft = int(remainingSecs / 3600 - dLeft * 24)
        mLeft = int(remainingSecs / 60 - hLeft * 60 - dLeft * 1440)
        sLeft = int(remainingSecs - hLeft * 3600 - mLeft * 60 - dLeft * 86400)
        self.log.info(
            "Storing proxy for credentials %s (%d:%02d:%02d:%02d left)" % (proxyIdentityDN, dLeft, hLeft, mLeft, sLeft)
        )

        try:
            sUserDN = self._escapeString(userDN)["Value"]
            sTable = "ProxyDB_CleanProxies"
        except KeyError:
            return S_ERROR("Cannot escape DN")
        # Check what we have already got in the repository
        cmd = "SELECT TIMESTAMPDIFF( SECOND, UTC_TIMESTAMP(), ExpirationTime ), Pem "
        cmd += "FROM `%s` WHERE UserDN=%s " % (sTable, sUserDN)
        result = self._query(cmd)
        if not result["OK"]:
            return result

        # Check if there is a previous ticket for the DN
        data = result["Value"]
        sqlInsert = True
        if data:
            sqlInsert = False
            pem = data[0][1]
            if pem:
                remainingSecsInDB = data[0][0]
                if remainingSecs <= remainingSecsInDB:
                    self.log.info(
                        "Proxy stored is longer than uploaded, omitting.",
                        "%s in uploaded, %s in db" % (remainingSecs, remainingSecsInDB),
                    )
                    return S_OK()

        pemChain = chain.dumpAllToString()["Value"]
        dValues = {
            "UserName": self._escapeString(userName)["Value"],
            "UserDN": sUserDN,
            "Pem": self._escapeString(pemChain)["Value"],
            "ExpirationTime": "TIMESTAMPADD( SECOND, %d, UTC_TIMESTAMP() )" % int(remainingSecs),
        }
        dValues["ProxyProvider"] = "'%s'" % proxyProvider
        if sqlInsert:
            sqlFields = []
            sqlValues = []
            for key in dValues:
                sqlFields.append(key)
                sqlValues.append(dValues[key])
            cmd = "INSERT INTO `%s` ( %s ) VALUES ( %s )" % (sTable, ", ".join(sqlFields), ", ".join(sqlValues))
        else:
            sqlSet = []
            sqlWhere = []
            for k in dValues:
                if k in ("UserDN", "ProxyProvider"):
                    sqlWhere.append("%s = %s" % (k, dValues[k]))
                else:
                    sqlSet.append("%s = %s" % (k, dValues[k]))
            cmd = "UPDATE `%s` SET %s WHERE %s" % (sTable, ", ".join(sqlSet), " AND ".join(sqlWhere))

        self.logAction("store proxy", userDN, proxyProvider, userDN, proxyProvider)
        return self._update(cmd)

    def purgeExpiredProxies(self, sendNotifications=True):
        """Purge expired requests from the db

        :param boolean sendNotifications: if need to send notification

        :return: S_OK(int)/S_ERROR() -- int is number of purged expired proxies
        """
        purged = 0
        for tableName in ("ProxyDB_CleanProxies", "ProxyDB_Proxies", "ProxyDB_VOMSProxies"):
            cmd = "DELETE FROM `%s` WHERE ExpirationTime < UTC_TIMESTAMP()" % tableName
            result = self._update(cmd)
            if not result["OK"]:
                return result
            purged += result["Value"]
            self.log.info("Purged %s expired proxies from %s" % (result["Value"], tableName))
        if sendNotifications:
            result = self.sendExpirationNotifications()
            if not result["OK"]:
                return result
        return S_OK(purged)

    def deleteProxy(self, userDN, userGroup=None, proxyProvider=None):
        """Remove proxy of the given user from the repository

        :param str userDN: user DN
        :param str userGroup: DIRAC group
        :param str proxyProvider: proxy provider name

        :return: S_OK()/S_ERROR()
        """
        try:
            userDN = self._escapeString(userDN)["Value"]
            if userGroup:
                userGroup = self._escapeString(userGroup)["Value"]
            if proxyProvider:
                proxyProvider = self._escapeString(proxyProvider)["Value"]
        except KeyError:
            return S_ERROR("Invalid DN or group or proxy provider")
        errMsgs = []
        req = "DELETE FROM `%%s` WHERE UserDN=%s" % userDN
        if proxyProvider or not userGroup:
            result = self._update(
                "%s %s" % (req % "ProxyDB_CleanProxies", proxyProvider and "AND ProxyProvider=%s" % proxyProvider or "")
            )
            if not result["OK"]:
                errMsgs.append(result["Message"])
        for table in ["ProxyDB_Proxies", "ProxyDB_VOMSProxies"]:
            result = self._update("%s %s" % (req % table, userGroup and "AND UserGroup=%s" % userGroup or ""))
            if not result["OK"]:
                if result["Message"] not in errMsgs:
                    errMsgs.append(result["Message"])
        if errMsgs:
            return S_ERROR(", ".join(errMsgs))
        return result

    def __getPemAndTimeLeft(self, userDN, userGroup=None, vomsAttr=None, proxyProvider=None):
        """Get proxy from database

        :param str userDN: user DN
        :param str userGroup: requested DIRAC group
        :param str vomsAttr: VOMS name
        :param str proxyProvider: proxy provider name

        :return: S_OK(tuple)/S_ERROR() -- tuple contain proxy as string and remaining seconds
        """
        try:
            sUserDN = self._escapeString(userDN)["Value"]
            if userGroup:
                sUserGroup = self._escapeString(userGroup)["Value"]
            if vomsAttr:
                sVomsAttr = self._escapeString(vomsAttr)["Value"]
        except KeyError:
            return S_ERROR("Invalid DN or Group")
        if proxyProvider:
            sTable = "`ProxyDB_CleanProxies`"
        elif not vomsAttr:
            sTable = "`ProxyDB_Proxies`"
        else:
            sTable = "`ProxyDB_VOMSProxies`"
        cmd = "SELECT Pem, TIMESTAMPDIFF( SECOND, UTC_TIMESTAMP(), ExpirationTime ) from %s " % sTable
        cmd += "WHERE UserDN=%s AND TIMESTAMPDIFF( SECOND, UTC_TIMESTAMP(), ExpirationTime ) > 0" % (sUserDN)
        if proxyProvider:
            cmd += ' AND ProxyProvider="%s"' % proxyProvider
        else:
            if userGroup:
                cmd += " AND UserGroup=%s" % sUserGroup
            if vomsAttr:
                cmd += " AND VOMSAttr=%s" % sVomsAttr
        retVal = self._query(cmd)
        if not retVal["OK"]:
            return retVal
        data = retVal["Value"]
        for record in data:
            if record[0]:
                if proxyProvider:
                    chain = X509Chain()
                    result = chain.loadProxyFromString(record[0])
                    if not result["OK"]:
                        return result
                    result = chain.generateProxyToString(record[1], diracGroup=userGroup, rfc=True)
                    if not result["OK"]:
                        return result
                    return S_OK((result["Value"], record[1]))
                return S_OK((record[0], record[1]))
        if userGroup:
            userMask = "%s@%s" % (userDN, userGroup)
        else:
            userMask = userDN
        return S_ERROR(DErrno.EPROXYFIND, "%s has no proxy registered" % userMask)

    def renewFromMyProxy(self, userDN, userGroup, lifeTime=None, chain=None):
        """Renew proxy from MyProxy

        :param str userDN: user DN
        :param str userGroup: user group
        :param int lifeTime: needed proxy live time in a seconds
        :param X509Chain chain: proxy as chain

        :return: S_OK(X509Chain/S_ERROR()
        """
        if not lifeTime:
            lifeTime = 43200
        if not self.__useMyProxy:
            return S_ERROR("myproxy is disabled")
        # Get the chain
        if not chain:
            retVal = self.__getPemAndTimeLeft(userDN, userGroup)
            if not retVal["OK"]:
                return retVal
            pemData = retVal["Value"][0]
            chain = X509Chain()
            retVal = chain.loadProxyFromString(pemData)
            if not retVal["OK"]:
                return retVal

        originChainLifeTime = chain.getRemainingSecs()["Value"]
        maxMyProxyLifeTime = self.getMyProxyMaxLifeTime()
        # If we have a chain that's 0.8 of max mplifetime don't ask to mp
        if originChainLifeTime > maxMyProxyLifeTime * 0.8:
            self.log.error(
                "Skipping myproxy download",
                "user %s %s  chain has %s secs and requested %s secs"
                % (userDN, userGroup, originChainLifeTime, maxMyProxyLifeTime),
            )
            return S_OK(chain)

        lifeTime *= 1.3
        if lifeTime > maxMyProxyLifeTime:
            lifeTime = maxMyProxyLifeTime
        self.log.info("Renewing proxy from myproxy", "user %s %s for %s secs" % (userDN, userGroup, lifeTime))

        myProxy = MyProxy(server=self.getMyProxyServer())
        retVal = myProxy.getDelegatedProxy(chain, lifeTime)
        if not retVal["OK"]:
            return retVal
        mpChain = retVal["Value"]
        retVal = mpChain.getRemainingSecs()
        if not retVal["OK"]:
            return S_ERROR("Can't retrieve remaining secs from renewed proxy: %s" % retVal["Message"])
        mpChainSecsLeft = retVal["Value"]
        if mpChainSecsLeft < originChainLifeTime:
            self.log.info(
                "Chain downloaded from myproxy has less lifetime than the one stored in the db",
                "\n Downloaded from myproxy: %s secs\n Stored in DB: %s secs" % (mpChainSecsLeft, originChainLifeTime),
            )
            return S_OK(chain)
        retVal = mpChain.getDIRACGroup()
        if not retVal["OK"]:
            return S_ERROR("Can't retrieve DIRAC Group from renewed proxy: %s" % retVal["Message"])
        chainGroup = retVal["Value"]
        if chainGroup != userGroup:
            return S_ERROR("Mismatch between renewed proxy group and expected: %s vs %s" % (userGroup, chainGroup))
        retVal = self.__storeProxy(userDN, userGroup, mpChain)
        if not retVal["OK"]:
            self.log.error("Cannot store proxy after renewal", retVal["Message"])
        retVal = myProxy.getServiceDN()
        if not retVal["OK"]:
            hostDN = userDN
        else:
            hostDN = retVal["Value"]
        self.logAction("myproxy renewal", hostDN, "host", userDN, userGroup)
        return S_OK(mpChain)

    # WARN: this method will not be needed if CS section Users/<user>/DNProperties will be for every user
    # in this case will be used proxy providers that described there
    def __getPUSProxy(self, userDN, userGroup, requiredLifetime, requestedVOMSAttr=False):
        result = Registry.getGroupsForDN(userDN)
        if not result["OK"]:
            return result

        validGroups = result["Value"]
        if userGroup not in validGroups:
            return S_ERROR("Invalid group %s for user" % userGroup)

        voName = Registry.getVOForGroup(userGroup)
        if not voName:
            return S_ERROR("Can not determine VO for group %s" % userGroup)

        retVal = self.__getVOMSAttribute(userGroup, requestedVOMSAttr)
        if not retVal["OK"]:
            return retVal
        vomsAttribute = retVal["Value"]["attribute"]
        vomsVO = retVal["Value"]["VOMSVO"]

        puspServiceURL = Registry.getVOOption(voName, "PUSPServiceURL")
        if not puspServiceURL:
            return S_ERROR("Can not determine PUSP service URL for VO %s" % voName)

        user = userDN.split(":")[-1]

        puspURL = "%s?voms=%s:%s&proxy-renewal=false&disable-voms-proxy=false" "&rfc-proxy=true&cn-label=user:%s" % (
            puspServiceURL,
            vomsVO,
            vomsAttribute,
            user,
        )

        try:
            proxy = urlopen(puspURL).read()
        except Exception:
            return S_ERROR("Failed to get proxy from the PUSP server")

        chain = X509Chain()
        chain.loadChainFromString(proxy)
        chain.loadKeyFromString(proxy)

        result = chain.getCredentials()
        if not result["OK"]:
            return S_ERROR("Failed to get a valid PUSP proxy")
        credDict = result["Value"]
        if credDict["identity"] != userDN:
            return S_ERROR("Requested DN does not match the obtained one in the PUSP proxy")
        timeLeft = credDict["secondsLeft"]

        result = chain.generateProxyToString(timeLeft, diracGroup=userGroup)
        if not result["OK"]:
            return result
        proxyString = result["Value"]
        return S_OK((proxyString, timeLeft))

    def __generateProxyFromProxyProvider(self, userDN, proxyProvider):
        """Get proxy from proxy provider

        :param str userDN: user DN for what need to create proxy
        :param str proxyProvider: proxy provider name that will ganarete proxy

        :return: S_OK(dict)/S_ERROR() -- dict with remaining seconds, proxy as a string and as a chain
        """
        gLogger.info("Getting proxy from proxyProvider", '(for "%s" DN by "%s")' % (userDN, proxyProvider))
        result = ProxyProviderFactory().getProxyProvider(proxyProvider)
        if not result["OK"]:
            return result
        pp = result["Value"]
        result = pp.getProxy(userDN)
        if not result["OK"]:
            return result
        proxyStr = result["Value"]
        chain = X509Chain()
        result = chain.loadProxyFromString(proxyStr)
        if not result["OK"]:
            return result
        result = chain.getRemainingSecs()
        if not result["OK"]:
            return result
        remainingSecs = result["Value"]
        result = self.__storeProxy(userDN, chain, proxyProvider)
        if result["OK"]:
            return S_OK({"proxy": proxyStr, "chain": chain, "remainingSecs": remainingSecs})
        return result

    def __getProxyFromProxyProviders(self, userDN, userGroup, requiredLifeTime):
        """Generate new proxy from exist clean proxy or from proxy provider
        for use with userDN in the userGroup

        :param str userDN: user DN
        :param str userGroup: required group name
        :param int requiredLifeTime: required proxy live time in a seconds

        :return: S_OK(tuple)/S_ERROR() -- tuple contain proxy as string and remainig seconds
        """
        result = Registry.getGroupsForDN(userDN)
        if not result["OK"]:
            return S_ERROR("Cannot generate proxy: %s" % result["Message"])
        if userGroup not in result["Value"]:
            return S_ERROR("Cannot generate proxy: Invalid group %s for user" % userGroup)
        result = Registry.getProxyProvidersForDN(userDN)

        errMsgs = []
        if result["OK"]:
            providers = result["Value"]
            providers.append("Certificate")
            for proxyProvider in providers:
                self.log.verbose("Try to get proxy from ProxyDB_CleanProxies")
                result = self.__getPemAndTimeLeft(userDN, userGroup, proxyProvider=proxyProvider)
                if result["OK"] and (not requiredLifeTime or result["Value"][1] > requiredLifeTime):
                    return result
                if len(providers) == 1:
                    return S_ERROR('Cannot generate proxy: No proxy providers found for "%s"' % userDN)
                self.log.verbose("Try to generate proxy from %s proxy provider" % proxyProvider)
                result = self.__generateProxyFromProxyProvider(userDN, proxyProvider)
                if result["OK"]:
                    chain = result["Value"]["chain"]
                    remainingSecs = result["Value"]["remainingSecs"]
                    result = chain.generateProxyToString(remainingSecs, diracGroup=userGroup, rfc=True)
                    if result["OK"]:
                        return S_OK((result["Value"], remainingSecs))
                errMsgs.append('"%s": %s' % (proxyProvider, result["Message"]))

        return S_ERROR("Cannot generate proxy%s" % (errMsgs and ": " + ", ".join(errMsgs) or ""))

    def getProxy(self, userDN, userGroup, requiredLifeTime=None):
        """Get proxy string from the Proxy Repository for use with userDN
        in the userGroup

        :param str userDN: user DN
        :param str userGroup: required DIRAC group
        :param int requiredLifeTime: required proxy live time in a seconds

        :return: S_OK(tuple)/S_ERROR() -- tuple with proxy as chain and proxy live time in a seconds
        """
        # Test that group enable to download
        if not Registry.isDownloadableGroup(userGroup):
            return S_ERROR('"%s" group is disable to download.' % userGroup)

        # WARN: this block will not be needed if CS section Users/<user>/DNProperties will be for every user
        # in this case will be used proxy providers that described there
        # Get the Per User SubProxy if one is requested
        if isPUSPdn(userDN):
            result = self.__getPUSProxy(userDN, userGroup, requiredLifeTime)
            if not result["OK"]:
                return result
            pemData = result["Value"][0]
            timeLeft = result["Value"][1]
            chain = X509Chain()
            result = chain.loadProxyFromString(pemData)
            if not result["OK"]:
                return result
            return S_OK((chain, timeLeft))

        # Standard proxy is requested
        self.log.verbose("Try to get proxy from ProxyDB_Proxies")
        retVal = self.__getPemAndTimeLeft(userDN, userGroup)
        errMsg = "Can't get proxy%s: " % (requiredLifeTime and " for %s seconds" % requiredLifeTime or "")
        if not retVal["OK"]:
            errMsg += "%s, try to generate new" % retVal["Message"]
            retVal = self.__getProxyFromProxyProviders(userDN, userGroup, requiredLifeTime=requiredLifeTime)
        elif requiredLifeTime:
            if retVal["Value"][1] < requiredLifeTime and not self.__useMyProxy:
                errMsg += "Stored proxy is not long lived enough, try to generate new"
                retVal = self.__getProxyFromProxyProviders(userDN, userGroup, requiredLifeTime=requiredLifeTime)
        if not retVal["OK"]:
            return S_ERROR(DErrno.EPROXYFIND, "%s; %s" % (errMsg, retVal["Message"]))
        pemData = retVal["Value"][0]
        timeLeft = retVal["Value"][1]
        chain = X509Chain()
        result = chain.loadProxyFromString(pemData)
        if not retVal["OK"]:
            return S_ERROR("%s; %s" % (errMsg, retVal["Message"]))
        if self.__useMyProxy:
            if requiredLifeTime:
                if timeLeft < requiredLifeTime:
                    retVal = self.renewFromMyProxy(userDN, userGroup, lifeTime=requiredLifeTime, chain=chain)
                    if not retVal["OK"]:
                        return S_ERROR("%s; the proxy lifetime from MyProxy is less than required." % errMsg)
                    chain = retVal["Value"]

        # Proxy is invalid for some reason, let's delete it
        if not chain.isValidProxy()["OK"]:
            self.deleteProxy(userDN, userGroup)
            return S_ERROR(DErrno.EPROXYFIND, "%s@%s has no proxy registered" % (userDN, userGroup))
        return S_OK((chain, timeLeft))

    def __getVOMSAttribute(self, userGroup, requiredVOMSAttribute=False):
        """Get VOMS attribute for DIRAC group

        :param str userGroup: DIRAC group
        :param boolean requiredVOMSAttribute: VOMS attribute

        :return: S_OK(dict)/S_ERROR() -- dict contain attribute and VOMS VO
        """
        if requiredVOMSAttribute:
            return S_OK({"attribute": requiredVOMSAttribute, "VOMSVO": Registry.getVOMSVOForGroup(userGroup)})

        csVOMSMapping = Registry.getVOMSAttributeForGroup(userGroup)
        if not csVOMSMapping:
            return S_ERROR("No mapping defined for group %s in the CS" % userGroup)

        return S_OK({"attribute": csVOMSMapping, "VOMSVO": Registry.getVOMSVOForGroup(userGroup)})

    def getVOMSProxy(self, userDN, userGroup, requiredLifeTime=None, requestedVOMSAttr=None):
        """Get proxy string from the Proxy Repository for use with userDN
        in the userGroup

        :param str userDN: user DN
        :param str userGroup: required DIRAC group
        :param int requiredLifeTime: required proxy live time in a seconds
        :param str requestedVOMSAttr: VOMS attribute

        :return: S_OK(tuple)/S_ERROR() -- tuple with proxy as chain and proxy live time in a seconds
        """
        retVal = self.__getVOMSAttribute(userGroup, requestedVOMSAttr)
        if not retVal["OK"]:
            return retVal
        vomsAttr = retVal["Value"]["attribute"]
        vomsVO = retVal["Value"]["VOMSVO"]

        # Look in the cache
        retVal = self.__getPemAndTimeLeft(userDN, userGroup, vomsAttr)
        if retVal["OK"]:
            pemData = retVal["Value"][0]
            vomsTime = retVal["Value"][1]
            chain = X509Chain()
            retVal = chain.loadProxyFromString(pemData)
            if retVal["OK"]:
                retVal = chain.getRemainingSecs()
                if retVal["OK"]:
                    remainingSecs = retVal["Value"]
                    if requiredLifeTime and requiredLifeTime <= vomsTime and requiredLifeTime <= remainingSecs:
                        return S_OK((chain, min(vomsTime, remainingSecs)))

        if isPUSPdn(userDN):
            # Get the Per User SubProxy if one is requested
            result = self.__getPUSProxy(userDN, userGroup, requiredLifeTime, requestedVOMSAttr)
            if not result["OK"]:
                return result
            pemData = result["Value"][0]
            chain = X509Chain()
            result = chain.loadProxyFromString(pemData)
            if not result["OK"]:
                return result

        else:
            # Get the stored proxy and dress it with the VOMS extension
            retVal = self.getProxy(userDN, userGroup, requiredLifeTime)
            if not retVal["OK"]:
                return retVal
            chain, _secsLeft = retVal["Value"]

            vomsMgr = VOMS()
            attrs = vomsMgr.getVOMSAttributes(chain).get("Value") or [""]
            if attrs[0]:
                if vomsAttr != attrs[0]:
                    return S_ERROR(
                        "Stored proxy has already a different VOMS attribute %s than requested %s"
                        % (attrs[0], vomsAttr)
                    )
            else:
                retVal = vomsMgr.setVOMSAttributes(chain, vomsAttr, vo=vomsVO)
                if not retVal["OK"]:
                    return retVal
                chain = retVal["Value"]

        # We have got the VOMS proxy, store it into the cache
        result = self.__storeVOMSProxy(userDN, userGroup, vomsAttr, chain)
        if not result["OK"]:
            return result
        return S_OK((chain, result["Value"]))

    def __storeVOMSProxy(self, userDN, userGroup, vomsAttr, chain):
        """Store VOMS proxy

        :param str userDN: user DN
        :param str userGroup: DIRAC group
        :param str vomsAttr: VOMS attribute
        :param X509Chain() chain: proxy as chain

        :return: S_OK(str)/S_ERROR()
        """
        retVal = self._getConnection()
        if not retVal["OK"]:
            return retVal
        connObj = retVal["Value"]
        retVal1 = VOMS().getVOMSProxyInfo(chain, "actimeleft")
        retVal2 = VOMS().getVOMSProxyInfo(chain, "timeleft")
        if not retVal1["OK"]:
            return retVal1
        if not retVal2["OK"]:
            return retVal2
        try:
            vomsSecsLeft1 = int(retVal1["Value"].strip())
            vomsSecsLeft2 = int(retVal2["Value"].strip())
            vomsSecsLeft = min(vomsSecsLeft1, vomsSecsLeft2)
        except Exception as e:
            return S_ERROR("Can't parse VOMS time left: %s" % str(e))
        secsLeft = min(vomsSecsLeft, chain.getRemainingSecs()["Value"])
        pemData = chain.dumpAllToString()["Value"]
        result = Registry.getUsernameForDN(userDN)
        if not result["OK"]:
            userName = ""
        else:
            userName = result["Value"]
        try:
            sUserName = self._escapeString(userName)["Value"]
            sUserDN = self._escapeString(userDN)["Value"]
            sUserGroup = self._escapeString(userGroup)["Value"]
            sVomsAttr = self._escapeString(vomsAttr)["Value"]
            sPemData = self._escapeString(pemData)["Value"]
        except KeyError:
            return S_ERROR("Could not escape some data")
        cmd = (
            "REPLACE INTO `ProxyDB_VOMSProxies` ( UserName, UserDN, UserGroup, VOMSAttr, Pem, ExpirationTime ) VALUES "
        )
        cmd += "( %s, %s, %s, %s, %s, TIMESTAMPADD( SECOND, %d, UTC_TIMESTAMP() ) )" % (
            sUserName,
            sUserDN,
            sUserGroup,
            sVomsAttr,
            sPemData,
            secsLeft,
        )
        result = self._update(cmd, conn=connObj)
        if not result["OK"]:
            return result
        return S_OK(secsLeft)

    def getUsers(self, validSecondsLeft=0, userMask=None):
        """Get all the distinct users from the Proxy Repository. Optionally, only users
        with valid proxies within the given validity period expressed in seconds

        :param int validSecondsLeft: validity period expressed in seconds
        :param str userMask: user name that need to add to search filter

        :return: S_OK(list)/S_ERROR() -- list contain dicts with user name, DN, group
                                         expiration time, persistent flag
        """
        data = []
        sqlCond = []
        if validSecondsLeft:
            try:
                validSecondsLeft = int(validSecondsLeft)
            except ValueError:
                return S_ERROR("Seconds left has to be an integer")
            sqlCond.append("TIMESTAMPDIFF( SECOND, UTC_TIMESTAMP(), ExpirationTime ) > %d" % validSecondsLeft)

        if userMask:
            try:
                sUserName = self._escapeString(userMask)["Value"]
            except KeyError:
                return S_ERROR("Can't escape user name")
            sqlCond.append("UserName = %s" % sUserName)

        for table, fields in [
            ("ProxyDB_CleanProxies", ("UserName", "UserDN", "ExpirationTime")),
            ("ProxyDB_Proxies", ("UserName", "UserDN", "UserGroup", "ExpirationTime", "PersistentFlag")),
        ]:
            cmd = "SELECT %s FROM `%s`" % (", ".join(fields), table)
            if sqlCond:
                cmd += " WHERE %s" % " AND ".join(sqlCond)
            retVal = self._query(cmd)
            if not retVal["OK"]:
                return retVal
            for record in retVal["Value"]:
                record = list(record)
                if table == "ProxyDB_CleanProxies":
                    record.insert(2, "")
                    record.insert(4, False)
                data.append(
                    {
                        "Name": record[0],
                        "DN": record[1],
                        "group": record[2],
                        "expirationtime": record[3],
                        "persistent": record[4] == "True",
                    }
                )
        return S_OK(data)

    def getCredentialsAboutToExpire(self, requiredSecondsLeft, onlyPersistent=True):
        """Get credentials about to expire for MyProxy

        :param int requiredSecondsLeft: required seconds left
        :param boolean onlyPersistent: look records only with persistent flag

        :return: S_OK()/S_ERROR()
        """
        cmd = "SELECT UserDN, UserGroup, ExpirationTime, PersistentFlag FROM `ProxyDB_Proxies`"
        cmd += " WHERE TIMESTAMPDIFF( SECOND, ExpirationTime, UTC_TIMESTAMP() ) < %d and " % requiredSecondsLeft
        cmd += "TIMESTAMPDIFF( SECOND, ExpirationTime, UTC_TIMESTAMP() ) > 0"
        if onlyPersistent:
            cmd += " AND PersistentFlag = 'True'"
        return self._query(cmd)

    def setPersistencyFlag(self, userDN, userGroup, persistent=True):
        """Set the proxy PersistentFlag to the flag value

        :param str userDN: user DN
        :param str userGroup: group name
        :param boolean persistent: enable persistent flag

        :return: S_OK()/S_ERROR()
        """

        try:
            sUserDN = self._escapeString(userDN)["Value"]
            sUserGroup = self._escapeString(userGroup)["Value"]
        except KeyError:
            return S_ERROR("Can't escape something")
        if persistent:
            sqlFlag = "True"
        else:
            sqlFlag = "False"
        retVal = self._query(
            "SELECT PersistentFlag FROM `ProxyDB_Proxies` WHERE UserDN=%s AND UserGroup=%s" % (sUserDN, sUserGroup)
        )
        sqlInsert = True
        if retVal["OK"]:
            data = retVal["Value"]
            if len(data) > 0:
                sqlInsert = False
                if data[0][0] == sqlFlag:
                    return S_OK()
        if sqlInsert:
            # If it's not in the db and we're removing the persistency then do nothing
            if not persistent:
                return S_OK()
            result = Registry.getUsernameForDN(userDN)
            if not result["OK"]:
                self.log.error("setPersistencyFlag: Can not retrieve username for DN", userDN)
                return result
            try:
                sUserName = self._escapeString(result["Value"])["Value"]
            except KeyError:
                return S_ERROR("Can't escape user name")
            cmd = "INSERT INTO `ProxyDB_Proxies` ( UserName, UserDN, UserGroup, Pem, ExpirationTime, PersistentFlag ) "
            cmd += " VALUES( %s, %s, %s, '', UTC_TIMESTAMP(), 'True' )" % (sUserName, sUserDN, sUserGroup)
        else:
            cmd = "UPDATE `ProxyDB_Proxies` SET PersistentFlag='%s' WHERE UserDN=%s AND UserGroup=%s" % (
                sqlFlag,
                sUserDN,
                sUserGroup,
            )
        retVal = self._update(cmd)
        if not retVal["OK"]:
            return retVal
        return S_OK()

    def getProxiesContent(self, selDict, sortList, start=0, limit=0):
        """Get the contents of the db, parameters are a filter to the db

        :param dict selDict: selection dict that contain fields and their posible values
        :param dict sortList: dict with sorting fields
        :param int start: search limit start
        :param int start: search limit amount

        :return: S_OK(dict)/S_ERROR() -- dict contain fields, record list, total records
        """
        data = []
        sqlWhere = ["Pem is not NULL"]
        for table, fields in [
            ("ProxyDB_CleanProxies", ("UserName", "UserDN", "ExpirationTime")),
            ("ProxyDB_Proxies", ("UserName", "UserDN", "UserGroup", "ExpirationTime", "PersistentFlag")),
        ]:
            cmd = "SELECT %s FROM `%s`" % (", ".join(fields), table)
            for field in selDict:
                if field not in fields:
                    continue
                fVal = selDict[field]
                if isinstance(fVal, (dict, tuple, list)):
                    sqlWhere.append(
                        "%s in (%s)" % (field, ", ".join([self._escapeString(str(value))["Value"] for value in fVal]))
                    )
                else:
                    sqlWhere.append("%s = %s" % (field, self._escapeString(str(fVal))["Value"]))
            sqlOrder = []
            if sortList:
                for sort in sortList:
                    if len(sort) == 1:
                        sort = (sort, "DESC")
                    elif len(sort) > 2:
                        return S_ERROR("Invalid sort %s" % sort)
                    if sort[0] not in fields:
                        if table == "ProxyDB_CleanProxies" and sort[0] in ["UserGroup", "PersistentFlag"]:
                            continue
                        return S_ERROR("Invalid sorting field %s" % sort[0])
                    if sort[1].upper() not in ("ASC", "DESC"):
                        return S_ERROR("Invalid sorting order %s" % sort[1])
                    sqlOrder.append("%s %s" % (sort[0], sort[1]))
            if sqlWhere:
                cmd = "%s WHERE %s" % (cmd, " AND ".join(sqlWhere))
            if sqlOrder:
                cmd = "%s ORDER BY %s" % (cmd, ", ".join(sqlOrder))
            if limit:
                try:
                    start = int(start)
                    limit = int(limit)
                except ValueError:
                    return S_ERROR("start and limit have to be integers")
                cmd += " LIMIT %d,%d" % (start, limit)
            retVal = self._query(cmd)
            if not retVal["OK"]:
                return retVal
            for record in retVal["Value"]:
                record = list(record)
                if table == "ProxyDB_CleanProxies":
                    record.insert(2, "")
                    record.insert(4, False)
                record[4] = record[4] == "True"
                data.append(record)
        totalRecords = len(data)
        return S_OK({"ParameterNames": fields, "Records": data, "TotalRecords": totalRecords})

    def logAction(self, action, issuerDN, issuerGroup, targetDN, targetGroup):
        """Add an action to the log

        :param str action: proxy action
        :param str issuerDN: user DN of issuer
        :param str issuerGroup: DIRAC group of issuer
        :param str targetDN: user DN of target
        :param str targetGroup: DIRAC group of target

        :return: S_ERROR()
        """
        try:
            sAction = self._escapeString(action)["Value"]
            sIssuerDN = self._escapeString(issuerDN)["Value"]
            sIssuerGroup = self._escapeString(issuerGroup)["Value"]
            sTargetDN = self._escapeString(targetDN)["Value"]
            sTargetGroup = self._escapeString(targetGroup)["Value"]
        except KeyError:
            return S_ERROR("Can't escape from death")
        cmd = "INSERT INTO `ProxyDB_Log` ( Action, IssuerDN, IssuerGroup, TargetDN, TargetGroup, Timestamp ) VALUES "
        cmd += "( %s, %s, %s, %s, %s, UTC_TIMESTAMP() )" % (sAction, sIssuerDN, sIssuerGroup, sTargetDN, sTargetGroup)
        retVal = self._update(cmd)
        if not retVal["OK"]:
            self.log.error("Can't add a proxy action log: ", retVal["Message"])

    def purgeLogs(self):
        """Purge expired requests from the db

        :return: S_OK()/S_ERROR()
        """
        cmd = "DELETE FROM `ProxyDB_Log` WHERE TIMESTAMPDIFF( SECOND, Timestamp, UTC_TIMESTAMP() ) > 15552000"
        return self._update(cmd)

    def getLogsContent(self, selDict, sortList, start=0, limit=0):
        """
        Function to get the contents of the logs table
          parameters are a filter to the db
        """
        fields = ("Action", "IssuerDN", "IssuerGroup", "TargetDN", "TargetGroup", "Timestamp")
        cmd = "SELECT %s FROM `ProxyDB_Log`" % ", ".join(fields)
        if selDict:
            qr = []
            if "beforeDate" in selDict:
                qr.append("Timestamp < %s" % self._escapeString(selDict["beforeDate"])["Value"])
                del selDict["beforeDate"]
            if "afterDate" in selDict:
                qr.append("Timestamp > %s" % self._escapeString(selDict["afterDate"])["Value"])
                del selDict["afterDate"]
            for field in selDict:
                qr.append(
                    "(%s)"
                    % " OR ".join(
                        ["%s=%s" % (field, self._escapeString(str(value))["Value"]) for value in selDict[field]]
                    )
                )
            whereStr = " WHERE %s" % " AND ".join(qr)
            cmd += whereStr
        else:
            whereStr = ""
        if sortList:
            cmd += " ORDER BY %s" % ", ".join(["%s %s" % (sort[0], sort[1]) for sort in sortList])
        if limit:
            cmd += " LIMIT %d,%d" % (start, limit)
        retVal = self._query(cmd)
        if not retVal["OK"]:
            return retVal
        data = retVal["Value"]
        totalRecords = len(data)
        cmd = "SELECT COUNT( Timestamp ) FROM `ProxyDB_Log`"
        cmd += whereStr
        retVal = self._query(cmd)
        if retVal["OK"]:
            totalRecords = retVal["Value"][0][0]
        return S_OK({"ParameterNames": fields, "Records": data, "TotalRecords": totalRecords})

    def generateToken(self, requesterDN, requesterGroup, numUses=1, lifeTime=0, retries=10):
        """Generate and return a token and the number of uses for the token

        :param str requesterDN: DN of requester
        :param str requesterGroup: DIRAC group of requester
        :param int numUses: number of uses
        :param int lifeTime: proxy live time in a seconds
        :param int retries: number of retries

        :return: S_OK(tuple)/S_ERROR() -- tuple with token and number of uses
        """
        if not lifeTime:
            lifeTime = gConfig.getValue("/DIRAC/VOPolicy/TokenLifeTime", self.__defaultTokenLifetime)
        maxUses = gConfig.getValue("/DIRAC/VOPolicy/TokenMaxUses", self.__defaultTokenMaxUses)
        numUses = max(1, min(numUses, maxUses))
        m = hashlib.md5()
        rndData = "%s.%s.%s.%s" % (time.time(), random.random(), numUses, lifeTime)
        m.update(rndData.encode())
        token = m.hexdigest()
        fieldsSQL = ", ".join(("Token", "RequesterDN", "RequesterGroup", "ExpirationTime", "UsesLeft"))
        valuesSQL = ", ".join(
            (
                self._escapeString(token)["Value"],
                self._escapeString(requesterDN)["Value"],
                self._escapeString(requesterGroup)["Value"],
                "TIMESTAMPADD( SECOND, %d, UTC_TIMESTAMP() )" % int(lifeTime),
                str(numUses),
            )
        )

        insertSQL = "INSERT INTO `ProxyDB_Tokens` ( %s ) VALUES ( %s )" % (fieldsSQL, valuesSQL)
        result = self._update(insertSQL)
        if result["OK"]:
            return S_OK([token, numUses])
        if result["Message"].find("uplicate entry") > -1:
            if retries:
                return self.generateToken(numUses, lifeTime, retries - 1)
            return S_ERROR("Max retries reached for token generation. Aborting")
        return result

    def purgeExpiredTokens(self):
        """Purge expired tokens from the db

        :return: S_OK(boolean)/S_ERROR()
        """
        delSQL = "DELETE FROM `ProxyDB_Tokens` WHERE ExpirationTime < UTC_TIMESTAMP() OR UsesLeft < 1"
        return self._update(delSQL)

    def useToken(self, token, requesterDN, requesterGroup):
        """Uses of token count

        :param str token: token
        :param str requesterDN: DN of requester
        :param str requesterGroup: DIRAC group of requester

        :return: S_OK(boolean)/S_ERROR()
        """
        sqlCond = " AND ".join(
            (
                "UsesLeft > 0",
                "Token=%s" % self._escapeString(token)["Value"],
                "RequesterDN=%s" % self._escapeString(requesterDN)["Value"],
                "RequesterGroup=%s" % self._escapeString(requesterGroup)["Value"],
                "ExpirationTime >= UTC_TIMESTAMP()",
            )
        )
        updateSQL = "UPDATE `ProxyDB_Tokens` SET UsesLeft = UsesLeft - 1 WHERE %s" % sqlCond
        result = self._update(updateSQL)
        if not result["OK"]:
            return result
        return S_OK(result["Value"] > 0)

    def __cleanExpNotifs(self):
        """Clean expired notifications from the db

        :return: S_OK()/S_ERROR()
        """
        cmd = "DELETE FROM `ProxyDB_ExpNotifs` WHERE ExpirationTime < UTC_TIMESTAMP()"
        return self._update(cmd)

    # FIXME: Add clean proxy
    def sendExpirationNotifications(self):
        """Send notification about expiration

        :return: S_OK(list)/S_ERROR() -- tuple list of user DN, group and proxy left time
        """
        result = self.__cleanExpNotifs()
        if not result["OK"]:
            return result
        cmd = "SELECT UserDN, UserGroup, LifeLimit FROM `ProxyDB_ExpNotifs`"
        result = self._query(cmd)
        if not result["OK"]:
            return result
        notifDone = dict([((row[0], row[1]), row[2]) for row in result["Value"]])
        notifLimits = sorted([int(x) for x in self.getCSOption("NotificationTimes", ProxyDB.NOTIFICATION_TIMES)])
        sqlSel = "UserDN, UserGroup, TIMESTAMPDIFF( SECOND, UTC_TIMESTAMP(), ExpirationTime )"
        sqlCond = "TIMESTAMPDIFF( SECOND, UTC_TIMESTAMP(), ExpirationTime ) < %d" % max(notifLimits)
        cmd = "SELECT %s FROM `ProxyDB_Proxies` WHERE %s" % (sqlSel, sqlCond)
        result = self._query(cmd)
        if not result["OK"]:
            return result
        pilotProps = (Properties.GENERIC_PILOT, Properties.PILOT)
        data = result["Value"]
        sent = []
        for row in data:
            userDN, group, lTime = row
            # If it's a pilot proxy, skip it
            if Registry.groupHasProperties(group, pilotProps):
                continue
            # IF it dosn't hace the auto upload proxy, skip it
            if not Registry.getGroupOption(group, "AutoUploadProxy", False):
                continue
            notKey = (userDN, group)
            for notifLimit in notifLimits:
                if notifLimit < lTime:
                    # Not yet in this notification limit
                    continue
                if notKey in notifDone and notifDone[notKey] <= notifLimit:
                    # Already notified for this notification limit
                    break
                if not self._notifyProxyAboutToExpire(userDN, lTime):
                    # Cannot send notification, retry later
                    break
                try:
                    sUserDN = self._escapeString(userDN)["Value"]
                    sGroup = self._escapeString(group)["Value"]
                except KeyError:
                    return S_ERROR("OOPS")
                if notKey not in notifDone:
                    values = "( %s, %s, %d, TIMESTAMPADD( SECOND, %s, UTC_TIMESTAMP() ) )" % (
                        sUserDN,
                        sGroup,
                        notifLimit,
                        lTime,
                    )
                    cmd = (
                        "INSERT INTO `ProxyDB_ExpNotifs` ( UserDN, UserGroup, LifeLimit, ExpirationTime ) VALUES %s"
                        % values
                    )
                    result = self._update(cmd)
                    if not result["OK"]:
                        gLogger.error("Could not mark notification as sent", result["Message"])
                else:
                    values = "LifeLimit = %d, ExpirationTime = TIMESTAMPADD( SECOND, %s, UTC_TIMESTAMP() )" % (
                        notifLimit,
                        lTime,
                    )
                    cmd = "UPDATE `ProxyDB_ExpNotifs` SET %s WHERE UserDN = %s AND UserGroup = %s" % (
                        values,
                        sUserDN,
                        sGroup,
                    )
                    result = self._update(cmd)
                    if not result["OK"]:
                        gLogger.error("Could not mark notification as sent", result["Message"])
                sent.append((userDN, group, lTime))
                notifDone[notKey] = notifLimit
        return S_OK(sent)

    def _notifyProxyAboutToExpire(self, userDN, lTime):
        """Send notification mail about to expire

        :param str userDN: user DN
        :param int lTime: left proxy live time in a seconds

        :return: boolean
        """
        result = Registry.getUsernameForDN(userDN)
        if not result["OK"]:
            return False
        userName = result["Value"]
        userEMail = Registry.getUserOption(userName, "Email", "")
        if not userEMail:
            gLogger.error("Could not discover user email", userName)
            return False
        daysLeft = int(lTime / 86400)
        msgSubject = "Your proxy uploaded to DIRAC will expire in %d days" % daysLeft
        msgBody = """\
Dear %s,

  The proxy you uploaded to DIRAC will expire in aproximately %d days. The proxy
  information is:

  DN:    %s

  If you have been issued different certificate, please make sure you have a
  proxy uploaded with that certificate.

Cheers,
 DIRAC's Proxy Manager
""" % (
            userName,
            daysLeft,
            userDN,
        )
        fromAddr = self._mailFrom
        result = self.__notifClient.sendMail(userEMail, msgSubject, msgBody, fromAddress=fromAddr)
        if not result["OK"]:
            gLogger.error("Could not send email", result["Message"])
            return False
        return True



""" DIRAC Encoding utilities based on json
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import datetime
import json


# Describes the way date time will be transmetted
# We do not keep miliseconds
DATETIME_DEFAULT_FORMAT = "%Y-%m-%d %H:%M:%S"
DATETIME_DEFAULT_DATE_FORMAT = "%Y-%m-%d"


class JSerializable(object):
    """
    Base class to define a serializable object by DIRAC.

    An object that ought to be serialized throught DISET shoud:
      * inherit from this class
      * define the _attrToSerialize list as class member. It is a list of
        strings containing the name of the attributes that should be serialized
      * have a constructor that takes no arguments, or only keywords arguments

    Exemple:

      class Serializable(JSerializable):

        _attrToSerialize = ['myAttr']

        def __init__(self, myAttr = None):
          self.myAttr = myAttr

    Limitations:
      * This will not work for classes defined inside classes. The class definition shoud be
        visible from the global scope
      * Class attributes cannot be serialized as such. They are converted to instance attributes.


    """

    def _toJSON(self):
        """Translates the objct into a dictionary.
        It is meant to be called by JSONDecoder only.

        It relies on the attribute _attrToSerialize to know which attributes to
        serialize.

        The returned dictionary contains the attributes serialized as well as
        hints for reconstructing the object upon receive.

        :raises TypeError: If the object is not serializable (no _attrToSerialize defined)

        :returns: a dictionary representing the object


        """

        # If the object does not have _attrToSerialize defined
        # Raise TypeError
        if not hasattr(self, "_attrToSerialize"):
            raise TypeError("Object not serializable. _attrToSerialize not defined")

        jsonData = {}

        # Store the class name and the module name
        jsonData["__dCls"] = self.__class__.__name__
        jsonData["__dMod"] = self.__module__

        # self._attrToSerialize will be defined by the child class
        for attr in self._attrToSerialize:  # pylint: disable=no-member
            # If an argument to serialize is not defined,
            # we continue. This is handy for arguments that
            # are defined dynamicaly ,like SQLAlchemy does.
            if not hasattr(self, attr):
                continue
            attrValue = getattr(self, attr)
            if attrValue is not None:
                jsonData[attr] = attrValue

        return jsonData


class DJSONEncoder(json.JSONEncoder):
    """This custom encoder is to add support to json for
    tuple, datetime, and any object inheriting from JSerializable
    """

    def default(self, obj):  # pylint: disable=method-hidden
        """Add supports for datetime and JSerializable class to default json

        :param obj: object to serialize

        :return: json string of the serialized objects

        """

        # If we have a datetime object, dumps its string representation
        if isinstance(obj, datetime.datetime):
            return {"__dCls": "dt", "obj": obj.strftime(DATETIME_DEFAULT_FORMAT)}
        elif isinstance(obj, datetime.date):
            return {"__dCls": "date", "obj": obj.strftime(DATETIME_DEFAULT_DATE_FORMAT)}
        # if the object inherits from JSJerializable, try to serialize it
        elif isinstance(obj, JSerializable):
            return obj._toJSON()  # pylint: disable=protected-access

        # otherwise, let the parent do
        return super(DJSONEncoder, self).default(obj)


class DJSONDecoder(json.JSONDecoder):
    """This custom decoder is to add support to json for
    tuple, datetime, and any object inheriting from JSerializable
    """

    def __init__(self, *args, **kargs):
        """
        Init method needed in order to give the object_hook to have special
        deserialization method.

        """
        super(DJSONDecoder, self).__init__(object_hook=self.dict_to_object, *args, **kargs)

    @staticmethod
    def dict_to_object(dataDict):
        """Convert the dictionary into an object.
        Adds deserialization support for datetype and JSerializable

        :param dataDict: json dictionary representing the data

        :returns: deserialized object
        """

        className = dataDict.pop("__dCls", None)

        # If the class is of type dt (datetime)
        if className == "dt":
            return datetime.datetime.strptime(dataDict["obj"], DATETIME_DEFAULT_FORMAT)
        elif className == "date":
            return datetime.datetime.strptime(dataDict["obj"], DATETIME_DEFAULT_DATE_FORMAT).date()
        elif className:
            import importlib

            # Get the module
            modName = dataDict.pop("__dMod")

            # Load the module
            mod = importlib.import_module(modName)
            # import the class
            cl = getattr(mod, className)
            # Instantiate the object
            obj = cl()

            # Set each attribute
            for attrName, attrValue in dataDict.items():
                # If the value is None, do not set it
                # This is needed to play along well with SQLalchemy
                if attrValue is None:
                    continue

                setattr(obj, attrName, attrValue)

            return obj

        # If we do not know how to serialize, just return the dictionary
        return dataDict


def encode(inData):
    """Encode the input data into a JSON string

    :param inData: anything that can be serialized.
                   Namely, anything that can be serialized by standard json package,
                   datetime object, tuples,  and any class that inherits from JSerializable

    :return: a json string
    """
    return json.dumps(inData, cls=DJSONEncoder)


def decode(encodedData):
    """Decode the json encoded string

    :param encodedData: json encoded string

    :return: the decoded objects, encoded object length

    Arguably, the length of the encodedData is useless,
    but it is for compatibility
    """
    return json.loads(encodedData, cls=DJSONDecoder), len(encodedData)


def strToIntDict(inDict):
    """Because JSON will transform dict with int keys to str keys,
    this utility method is just to cast it back.
    This shows useful in cases when sending dict indexed on
    jobID or requestID for example

    :param inDict: dictionary with strings as keys e.g. {'1': 1, '2': 2}

    :returns: dictionary with int as keys e.g. {1: 1, 2: 2}

    """

    return {int(key): value for key, value in inDict.items()}



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import importlib

import six

from DIRAC import S_OK
from DIRAC.Core.Utilities.Extensions import extensionsByPriority


def getCurrentVersion():
    """Get a string corresponding to the current version of the DIRAC package and all the installed
    extension packages
    """
    for ext in extensionsByPriority():
        try:
            return S_OK(importlib.import_module(ext).version)
        except (ImportError, AttributeError):
            pass


def getVersion():
    """Get a dictionary corresponding to the current version of the DIRAC package and all the installed
    extension packages
    """
    vDict = {"Extensions": {}}
    for ext in extensionsByPriority():
        if six.PY2:
            try:
                version = importlib.import_module(ext).version
            except (ImportError, AttributeError):
                continue
            if ext.endswith("DIRAC") and ext != "DIRAC":
                ext = ext[: -len("DIRAC")]
        else:
            from importlib.metadata import version as get_version  # pylint: disable=import-error,no-name-in-module

            version = get_version(ext)

        vDict["Extensions"][ext] = version
    return S_OK(vDict)



"""Job Information"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from pprint import pformat
import six
from six.moves import zip_longest

from DIRAC import gLogger
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.Core.Utilities.List import fromChar

ASSIGNEDSTATES = ["Assigned", "Processed"]

LOG = gLogger.getSubLogger(__name__)
__RCSID__ = "$Id$"


class TaskInfoException(Exception):
    """Exception when the task info is not recoverable"""

    def __init__(self, message):
        super(TaskInfoException, self).__init__(message)


class JobInfo(object):
    """hold information about jobs"""

    def __init__(self, jobID, status, tID, tType=None):
        self.tID = int(tID)
        self.tType = tType
        self.jobID = int(jobID)
        self.status = status

        # information related to input files
        self.inputFiles = []
        self.inputFileStatus = []  # from the filecatalog
        self.transFileStatus = []  # from the transformation
        self.inputFilesExist = []
        self.errorCounts = []
        self.otherTasks = []

        self.outputFiles = []
        self.outputFileStatus = []

        self.taskID = None
        self.pendingRequest = False

        self.maxResetCounter = Operations().getValue("Transformations/FilesMaxResetCounter", 10)

    def __str__(self):
        info = "%d: %s" % (self.jobID, self.status)
        if self.tID and self.taskID:
            info += " %s Transformation: %d -- %d " % (self.tType, self.tID, self.taskID)
        if self.otherTasks:
            info += " (Last task %s)" % self.otherTasks
        if self.inputFiles:
            ifInfo = [
                "<<< %s (%s, %s, Errors %s)" % _
                for _ in zip_longest(self.inputFiles, self.inputFilesExist, self.transFileStatus, self.errorCounts)
            ]
            info += "\n".join(ifInfo)
        if self.outputFiles:
            info += "\n::: OutputFiles: "
            efInfo = ["%s (%s)" % _ for _ in zip_longest(self.outputFiles, self.outputFileStatus)]
            info += ", ".join(efInfo)
        if self.pendingRequest:
            info += "\n PENDING REQUEST IGNORE THIS JOB!!!"
        else:
            info += "\n No Pending Requests"

        return info

    def allFilesExist(self):
        """Check if all files exists, if there are no output files, return False."""
        return bool(self.outputFileStatus) and all("Exists" in status for status in self.outputFileStatus)

    def allFilesMissing(self):
        """Check if all files are missing, if not output files, return False"""
        return bool(self.outputFileStatus) and all("Missing" in status for status in self.outputFileStatus)

    def someFilesMissing(self):
        """check if some files are missing and therefore some files exist"""
        return bool(self.outputFileStatus) and not (self.allFilesExist() or self.allFilesMissing())

    def allInputFilesExist(self):
        """check if all input files exists"""
        if not self.inputFileStatus:
            return False
        return all("Exists" in status for status in self.inputFileStatus)

    def allInputFilesMissing(self):
        """check if all input files are missing"""
        if not self.inputFileStatus:
            return False
        return all("Missing" in status for status in self.inputFileStatus)

    def someInputFilesMissing(self):
        """check if some input files are missing and therefore some files exist"""
        if not self.inputFileStatus:
            return False
        return not (self.allInputFilesExist() or self.allInputFilesMissing())

    def allFilesProcessed(self):
        """Check if all input files are processed."""
        return bool(self.transFileStatus) and all("Processed" in status for status in self.transFileStatus)

    def allFilesAssigned(self):
        """Check if all input files are assigned."""
        return bool(self.transFileStatus) and all(status in ASSIGNEDSTATES for status in self.transFileStatus)

    def allTransFilesDeleted(self):
        """Check if all input files are deleted in the Transformation System."""
        return bool(self.transFileStatus) and all(status == "Deleted" for status in self.transFileStatus)

    def checkErrorCount(self):
        """Check if any file is above Operations/Transformations/FilesMaxResetCounter error count."""
        return any(errorCount > self.maxResetCounter for errorCount in self.errorCounts)

    def getJobInformation(self, diracAPI, jobMon, jdlOnly=False):
        """Get all the information for the job.

        The InputData, TaskID, OutputData can either be taken from properly filled JDL or

        * inputData from jobMonitor getInputData
        * TaskID from the name of The job via jobMonitor getJobAttribute JobName
        * ProductionOutputData: from jobMonitor getJobParameter ProductionOutputData

        This would be faster if we could do bulk calls for all of these
        """
        if not jdlOnly:
            # this is actually slower than just getting the jdl, because getting the jdl is one service call
            # this is three service calls to three different DBs
            resInputData = jobMon.getInputData(self.jobID)
            if resInputData["OK"]:
                self.inputFiles = resInputData["Value"]
            resName = jobMon.getJobAttribute(self.jobID, "JobName")
            if resName["OK"] and "_" in resName["Value"]:
                self.__getTaskID(resName["Value"])
            resOutput = jobMon.getJobParameter(self.jobID, "ProductionOutputData")
            if resOutput["OK"]:
                self.outputFiles = fromChar(resOutput["Value"].get("ProductionOutputData", ""))
                if not self.outputFiles:
                    LOG.verbose("Did not find outputFiles for", str(self))

        if not (self.inputFiles and self.outputFiles and self.taskID):
            LOG.verbose("Have to check JDL")
            jdlParameters = self.__getJDL(diracAPI)
            # get taskID from JobName, get inputfile(s) from DownloadInputdata
            self.__getOutputFiles(jdlParameters)
            self.__getTaskID(jdlParameters)
            if not self.inputFiles:
                self.__getInputFile(jdlParameters)

    def getTaskInfo(self, tasksDict, lfnTaskDict, withInputTypes):
        """extract the task information from the taskDict"""

        if not self.inputFiles and self.tType in withInputTypes:
            raise TaskInfoException("InputFiles is empty: %s" % str(self))

        # taskID not in tasksDict means another job has been assigned to the files
        if self.taskID in tasksDict:
            for taskDict in tasksDict[self.taskID]:
                self.transFileStatus.append(taskDict["Status"])
                self.errorCounts.append(taskDict["ErrorCount"])
                if taskDict["LFN"] not in self.inputFiles:
                    raise TaskInfoException(
                        "InputFiles do not agree: %s not in InputFiles: \n %s" % (taskDict["LFN"], str(self))
                    )
            return

        # taskID not in tasksDict means another job has been assigned to the files
        for inputLFN in self.inputFiles:
            try:
                for taskDict in tasksDict[lfnTaskDict[inputLFN]]:
                    self.otherTasks.append(lfnTaskDict[inputLFN])
                    self.transFileStatus.append(taskDict["Status"])
                    self.errorCounts.append(taskDict["ErrorCount"])
            except KeyError as ke:
                LOG.error("ERROR for key:", str(ke))
                LOG.error("Failed to get taskDict", "%s, %s: %s" % (self.taskID, self.inputFiles, pformat(lfnTaskDict)))
                raise

    def checkFileExistence(self, success):
        """check if input and outputfile still exist"""
        for lfn in self.inputFiles:
            if lfn in success and success[lfn]:
                self.inputFilesExist.append(True)
                self.inputFileStatus.append("Exists")
            elif lfn in success:
                self.inputFileStatus.append("Missing")
                self.inputFilesExist.append(False)
            else:
                self.inputFileStatus.append("Unknown")
                self.inputFilesExist.append(False)

        for lfn in self.outputFiles:
            if lfn in success and success[lfn]:
                self.outputFileStatus.append("Exists")
            elif lfn in success:
                self.outputFileStatus.append("Missing")
            else:
                self.outputFileStatus.append("Unknown")

    def __getJDL(self, diracAPI):
        """return jdlParameterDictionary for this job"""
        res = diracAPI.getJobJDL(int(self.jobID))
        if not res["OK"]:
            raise RuntimeError("Failed to get jobJDL: %s" % res["Message"])
        jdlParameters = res["Value"]
        return jdlParameters

    def __getOutputFiles(self, jdlParameters):
        """get the Production Outputfiles for the given Job"""
        lfns = jdlParameters.get("ProductionOutputData", [])
        if isinstance(lfns, six.string_types):
            lfns = [lfns]
        self.outputFiles = lfns

    def __getInputFile(self, jdlParameters):
        """get the Inputdata for the given job"""
        lfn = jdlParameters.get("InputData", None)
        if isinstance(lfn, six.string_types):
            self.inputFiles.append(lfn)
            return
        if isinstance(lfn, list):
            self.inputFiles = lfn
            return

    def __getTaskID(self, jdlParameters):
        """Get the taskID."""
        if isinstance(jdlParameters, six.string_types):
            self.taskID = int(jdlParameters.split("_")[1])
        if "TaskID" not in jdlParameters:
            return
        try:
            self.taskID = int(jdlParameters.get("TaskID", None))
        except ValueError:
            LOG.warn("*" * 80)
            LOG.warn("TaskID broken?: %r" % jdlParameters.get("TaskID", None))
            LOG.warn(self)
            LOG.warn("*" * 80)
            raise

    def setJobDone(self, tInfo):
        """mark job as done in wms and transformationsystem"""
        tInfo.setJobDone(self)

    def setJobFailed(self, tInfo):
        """mark job as failed in  wms and transformationsystem"""
        tInfo.setJobFailed(self)

    def setInputProcessed(self, tInfo):
        """mark input file as Processed"""
        tInfo.setInputProcessed(self)

    def setInputMaxReset(self, tInfo):
        """mark input file as MaxReset"""
        tInfo.setInputMaxReset(self)

    def setInputDeleted(self, tInfo):
        """mark input file as Deleted"""
        tInfo.setInputDeleted(self)

    def setInputUnused(self, tInfo):
        """mark input file as Unused"""
        tInfo.setInputUnused(self)

    def cleanOutputs(self, tInfo):
        """remove all job outputs"""
        tInfo.cleanOutputs(self)



"""
RMSMonitoring type is used to monitor behaviour pattern of requests executed by RequestManagementSystem
inside DataManagementSystem/Agent/RequestOperations.

Understanding the key fields:
'objectType': This field describes the type of object which can be Request, Operation, or File.
'operationType': This field contains the type of operation being performed for eg. MoveReplica, PutAndRegister, etc.
'status': This field contains the status of the operation performed i.e. Attempted, Failed, or Successful.
'objectID': This field will be RequestID if Request object / OperationID if Operation object.
'parentID': This field will be RequestID if Operation object / OperationID if File object.

Understanding the monitoring fields:
'nbObject': This field is used to describe the number of objects in question during the operation.

In order to enable 'RMSMonitoring' we need to set value of 'EnableRMSMonitoring' flag to 'yes/true' inside
'/Systems/RequestManagement/<Instance>/Agents/RequestExecutingAgent' of the 'cfg' file.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.MonitoringSystem.Client.Types.BaseType import BaseType


class RMSMonitoring(BaseType):
    """
    .. class:: RMSMonitoring
    """

    def __init__(self):
        """c'tor

        :param self: self reference
        """
        super(RMSMonitoring, self).__init__()

        self.keyFields = ["host", "objectType", "operationType", "status", "objectID", "parentID"]

        self.monitoringFields = ["nbObject"]

        self.addMapping(
            {
                "host": {"type": "keyword"},
                "objectType": {"type": "keyword"},
                "operationType": {"type": "keyword"},
                "status": {"type": "keyword"},
                "objectID": {"type": "long"},
                "parentID": {"type": "long"},
            }
        )

        self.dataToKeep = 86400 * 30  # we need to define...

        self.period = "month"
        self.checkType()



#!/bin/env python
"""
Show request given its ID, a jobID or a transformation and a task
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import datetime
import os
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


def convertDate(date):
    try:
        value = datetime.datetime.strptime(date, "%Y-%m-%d")
        return value
    except Exception:
        pass
    try:
        value = datetime.datetime.utcnow() - datetime.timedelta(hours=int(24 * float(date)))
    except Exception:
        from DIRAC import gLogger

        gLogger.fatal("Invalid date", date)
        value = None
    return value


@Script()
def main():
    Script.registerSwitch("", "Job=", "   JobID[,jobID2,...]")
    Script.registerSwitch("", "Transformation=", "   transformation ID")
    Script.registerSwitch("", "Tasks=", "      Associated to --Transformation, list of taskIDs")
    Script.registerSwitch("", "Verbose", "   Print more information")
    Script.registerSwitch("", "Terse", "   Only print request status")
    Script.registerSwitch("", "Full", "   Print full request content")
    Script.registerSwitch("", "Status=", "   Select all requests in a given status")
    Script.registerSwitch(
        "", "Since=", "      Associated to --Status, start date yyyy-mm-dd or nb of days (default= -one day"
    )
    Script.registerSwitch("", "Until=", "      Associated to --Status, end date (default= now")
    Script.registerSwitch("", "Maximum=", "      Associated to --Status, max number of requests ")
    Script.registerSwitch("", "Reset", "   Reset Failed files to Waiting if any")
    Script.registerSwitch("", "Force", "   Force reset even if not Failed")
    Script.registerSwitch(
        "", "All", "      (if --Status Failed) all requests, otherwise exclude irrecoverable failures"
    )
    Script.registerSwitch("", "FixJob", "   Set job Done if the request is Done")
    Script.registerSwitch("", "Cancel", "   Cancel the request")
    Script.registerSwitch("", "ListJobs", " List the corresponding jobs")
    Script.registerSwitch("", "TargetSE=", " Select request only if that SE is in the targetSEs")
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(
        (
            "file:     a file containing a list of requests (Comma-separated on each line)",
            "request:  a request ID or a unique request name",
        ),
        mandatory=False,
    )
    Script.registerArgument(["request:  a request ID or a unique request name"], mandatory=False)
    Script.parseCommandLine()

    import DIRAC
    from DIRAC import gLogger

    jobs = []
    requestID = 0
    transID = None
    taskIDs = None
    tasks = None
    requests = []
    full = False
    verbose = False
    status = None
    until = None
    since = None
    terse = False
    allR = False
    reset = False
    fixJob = False
    maxRequests = 999999999999
    cancel = False
    listJobs = False
    force = False
    targetSE = set()
    for switch in Script.getUnprocessedSwitches():
        if switch[0] == "Job":
            jobs = []
            job = "Unknown"
            try:
                for arg in switch[1].split(","):
                    if os.path.exists(arg):
                        with open(arg, "r") as fp:
                            lines = fp.readlines()
                        for line in lines:
                            for job in line.split(","):
                                jobs += [int(job.strip())]
                        gLogger.notice("Found %d jobs in file %s" % (len(jobs), arg))
                    else:
                        jobs.append(int(arg))
            except TypeError:
                gLogger.fatal("Invalid jobID", job)
        elif switch[0] == "Transformation":
            try:
                transID = int(switch[1])
            except Exception:
                gLogger.fatal("Invalid transID", switch[1])
        elif switch[0] == "Tasks":
            try:
                taskIDs = [int(task) for task in switch[1].split(",")]
            except Exception:
                gLogger.fatal("Invalid tasks", switch[1])
        elif switch[0] == "Full":
            full = True
        elif switch[0] == "Verbose":
            verbose = True
        elif switch[0] == "Terse":
            terse = True
        elif switch[0] == "All":
            allR = True
        elif switch[0] == "Reset":
            reset = True
        elif switch[0] == "Force":
            force = True
        elif switch[0] == "Status":
            status = switch[1].capitalize()
        elif switch[0] == "Since":
            since = convertDate(switch[1])
        elif switch[0] == "Until":
            until = convertDate(switch[1])
        elif switch[0] == "FixJob":
            fixJob = True
        elif switch[0] == "Cancel":
            cancel = True
        elif switch[0] == "ListJobs":
            listJobs = True
        elif switch[0] == "Maximum":
            try:
                maxRequests = int(switch[1])
            except Exception:
                pass
        elif switch[0] == "TargetSE":
            targetSE = set(switch[1].split(","))

    if reset and not force:
        status = "Failed"
    if fixJob:
        status = "Done"
    if terse:
        verbose = True
    if status:
        if not until:
            until = datetime.datetime.utcnow()
        if not since:
            since = until - datetime.timedelta(hours=24)
    from DIRAC.RequestManagementSystem.Client.ReqClient import ReqClient
    from DIRAC.RequestManagementSystem.Client.ReqClient import printRequest, recoverableRequest

    reqClient = ReqClient()
    if transID:
        if not taskIDs:
            gLogger.fatal("If Transformation is set, a list of Tasks should also be set")
            Script.showHelp(exitCode=2)
        # In principle, the task name is unique, so the request name should be unique as well
        # If ever this would not work anymore, we would need to use the transformationClient
        # to fetch the ExternalID
        requests = ["%08d_%08d" % (transID, task) for task in taskIDs]
        allR = True

    elif not jobs:
        requests = []
        # Get full list of arguments, with and without comma
        for arg in [x.strip() for arg in Script.getPositionalArgs() for x in arg.split(",")]:
            if os.path.exists(arg):
                lines = open(arg, "r").readlines()
                requests += [reqID.strip() for line in lines for reqID in line.split(",")]
                gLogger.notice("Found %d requests in file" % len(requests))
            else:
                requests.append(arg)
            allR = True
    else:
        res = reqClient.getRequestIDsForJobs(jobs)
        if not res["OK"]:
            gLogger.fatal("Error getting request for jobs", res["Message"])
            DIRAC.exit(2)
        if res["Value"]["Failed"]:
            gLogger.error("No request found for jobs %s" % ",".join(sorted(str(job) for job in res["Value"]["Failed"])))
        requests = sorted(res["Value"]["Successful"].values())
        if requests:
            allR = True
        else:
            DIRAC.exit(0)

    if status and not requests:
        allR = allR or status != "Failed"
        res = reqClient.getRequestIDsList([status], limit=maxRequests, since=since, until=until)

        if not res["OK"]:
            gLogger.error("Error getting requests:", res["Message"])
            DIRAC.exit(2)
        requests = [reqID for reqID, _st, updTime in res["Value"] if updTime > since and updTime <= until and reqID]
        gLogger.notice("Obtained %d requests %s between %s and %s" % (len(requests), status, since, until))
    if not requests:
        gLogger.notice("No request selected....")
        Script.showHelp(exitCode=2)
    okRequests = []
    warningPrinted = False
    jobIDList = []
    for reqID in requests:
        # We allow reqID to be the requestName if it is unique
        try:
            requestID = int(reqID)
        except ValueError:
            requestID = reqClient.getRequestIDForName(reqID)
            if not requestID["OK"]:
                gLogger.notice(requestID["Message"])
                continue
            requestID = requestID["Value"]

        request = reqClient.peekRequest(requestID)
        if not request["OK"]:
            gLogger.error(request["Message"])
            DIRAC.exit(-1)

        request = request["Value"]
        if not request:
            gLogger.error("no such request %s" % requestID)
            continue
        # If no operation as the targetSE, skip
        if targetSE:
            found = False
            for op in request:
                if op.TargetSE and targetSE.intersection(op.TargetSE.split(",")):
                    found = True
                    break
            if not found:
                continue
        # keep a list of jobIDs if requested
        if request.JobID and listJobs:
            jobIDList.append(request.JobID)

        if status and request.Status != status:
            gLogger.notice(
                "Request %s is not in requested status %s%s" % (reqID, status, " (cannot be reset)" if reset else "")
            )
            continue

        if fixJob and request.Status == "Done" and request.JobID:
            # The request is for a job and is Done, verify that the job is in the proper status
            result = reqClient.finalizeRequest(request.RequestID, request.JobID, useCertificates=False)
            if not result["OK"]:
                gLogger.error("Error finalizing job", result["Message"])
            else:
                gLogger.notice("Job %d updated to %s" % (request.JobID, result["Value"]))
            continue

        if cancel:
            if request.Status not in ("Done", "Failed"):
                ret = reqClient.cancelRequest(requestID)
                if not ret["OK"]:
                    gLogger.error("Error canceling request %s" % reqID, ret["Message"])
                else:
                    gLogger.notice("Request %s cancelled" % reqID)
            else:
                gLogger.notice("Request %s is in status %s, not cancelled" % (reqID, request.Status))

        elif allR or recoverableRequest(request):
            okRequests.append(str(requestID))
            if reset:
                gLogger.notice("============ Request %s =============" % requestID)
                ret = reqClient.resetFailedRequest(requestID, allR=allR)
                if not ret["OK"]:
                    gLogger.error("Error resetting request %s" % requestID, ret["Message"])
            else:
                if len(requests) > 1:
                    gLogger.notice("\n===================================")
                dbStatus = reqClient.getRequestStatus(requestID).get("Value", "Unknown")
                printRequest(request, status=dbStatus, full=full, verbose=verbose, terse=terse)

    if listJobs:
        gLogger.notice("List of %d jobs:\n" % len(jobIDList), ",".join(str(jobID) for jobID in jobIDList))

    if status and okRequests:
        from DIRAC.Core.Utilities.List import breakListIntoChunks

        gLogger.notice("\nList of %d selected requests:" % len(okRequests))
        for reqs in breakListIntoChunks(okRequests, 100):
            gLogger.notice(",".join(reqs))


if __name__ == "__main__":
    main()



""" Class to manage connections for the Message Queue resources.
    Also, set of 'private' helper functions to access and modify the message queue connection storage.
    They are ment to be used only internally by the MQConnectionManager, which should
    assure thread-safe access to it and standard S_OK/S_ERROR error handling.
    MQConnection storage is a dict structure that contains the MQ connections used and reused for
    producer/consumer communication. Example structure::

      {
        mardirac3.in2p3.fr: {'MQConnector':StompConnector, 'destinations':{'/queue/test1':['consumer1', 'producer1'],
                                                                           '/queue/test2':['consumer1', 'producer1']}},
        blabal.cern.ch:     {'MQConnector':None,           'destinations':{'/queue/test2':['consumer2', 'producer2',]}}
      }
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Utilities.LockRing import LockRing
from DIRAC.Resources.MessageQueue.Utilities import getMQService
from DIRAC.Resources.MessageQueue.Utilities import getDestinationAddress
from DIRAC.Resources.MessageQueue.MQConnector import createMQConnector
from DIRAC.Core.Utilities.DErrno import EMQCONN


class MQConnectionManager(object):
    """Manages connections for the Message Queue resources in form of the interal connection storage."""

    def __init__(self, connectionStorage=None):
        self.log = gLogger.getSubLogger(self.__class__.__name__)
        self.__lock = None
        if connectionStorage:
            self.__connectionStorage = connectionStorage
        else:
            self.__connectionStorage = {}

    @property
    def lock(self):
        """Lock to assure thread-safe access to the internal connection storage."""
        if not self.__lock:
            self.__lock = LockRing().getLock(self.__class__.__name__, recursive=True)
        return self.__lock

    def startConnection(self, mqURI, params, messengerType):
        """Function adds or updates the MQ connection. If the connection
            does not exists, MQconnector is created and added.

        Args:
          mqURI(str):
          params(dict): parameters to initialize the MQConnector.
          messengerType(str): 'consumer' or 'producer'.
        Returns:
          S_OK/S_ERROR: with the value of the messenger Id in S_OK.
        """
        self.lock.acquire()
        try:
            conn = getMQService(mqURI)
            if self.__connectionExists(conn):
                return self.addNewMessenger(mqURI=mqURI, messengerType=messengerType)
            else:  # Connection does not exist so we create the connector and we add a new connection
                result = self.addNewMessenger(mqURI=mqURI, messengerType=messengerType)
                if not result["OK"]:
                    return result
                mId = result["Value"]
                result = self.createConnectorAndConnect(parameters=params)
                if not result["OK"]:
                    return result
                if self.__getConnector(conn):
                    return S_ERROR(EMQCONN, "The connector already exists!")
                self.__setConnector(conn, result["Value"])
                return S_OK(mId)
        finally:
            self.lock.release()

    def addNewMessenger(self, mqURI, messengerType):
        """Function updates the MQ connection by adding the messenger Id to the internal connection storage.
            Also the messengerId is chosen.
            messenger Id is set to the maximum existing value (or 0 no messengers are connected) + 1.
            messenger Id is calculated separately for consumers and producers

        Args:
          mqURI(str):
          messengerType(str): 'consumer' or 'producer'.
        Returns:
          S_OK: with the value of the messenger Id or S_ERROR if the messenger was not added,
                cause the same id already exists.
        """
        # 'consumer1' ->1
        # 'producer21' ->21
        def msgIdToInt(msgIds, msgType):
            return [int(m.replace(msgType, "")) for m in msgIds]

        # The messengerId is str e.g.  'consumer5' or 'producer3'

        def generateMessengerId(msgT):
            return msgT + str(max(msgIdToInt(self.__getAllMessengersIdWithType(msgT), msgT) or [0]) + 1)

        self.lock.acquire()
        try:
            conn = getMQService(mqURI)
            dest = getDestinationAddress(mqURI)
            mId = generateMessengerId(messengerType)
            if self.__addMessenger(conn, dest, mId):
                return S_OK(mId)
            return S_ERROR(EMQCONN, "Failed to update the connection: the messenger %s  already exists" % mId)
        finally:
            self.lock.release()

    def createConnectorAndConnect(self, parameters):
        result = createMQConnector(parameters=parameters)
        if not result["OK"]:
            return result
        connector = result["Value"]
        result = connector.setupConnection(parameters=parameters)
        if not result["OK"]:
            return result
        result = connector.connect()
        if not result["OK"]:
            return result
        return S_OK(connector)

    def disconnect(self, connector):
        if not connector:
            return S_ERROR(EMQCONN, "Failed to disconnect! Connector is None!")
        return connector.disconnect()

    def unsubscribe(self, connector, destination, messengerId):
        if not connector:
            return S_ERROR(EMQCONN, "Failed to unsubscribe! Connector is None!")
        return connector.unsubscribe(parameters={"destination": destination, "messengerId": messengerId})

    def getConnector(self, mqConnection):
        """Function returns MQConnector assigned to the mqURI.

        Args:
          mqConnection(str): connection name.
        Returns:
          S_OK/S_ERROR: with the value of the MQConnector in S_OK if not None
        """
        self.lock.acquire()
        try:
            connector = self.__getConnector(mqConnection)
            if not connector:
                return S_ERROR("Failed to get the MQConnector!")
            return S_OK(connector)
        finally:
            self.lock.release()

    def stopConnection(self, mqURI, messengerId):
        """Function 'stops' the connection for given messenger, which means
            it removes it from the messenger list. If this is the consumer, the
            unsubscribe() connector method is called. If it is the last messenger
            of this destination (queue or topic), then the destination is removed.
            If it is the last destination from this connection. The disconnect function
            is called and the connection is removed.

        Args:
          mqURI(str):
          messengerId(str): e.g. 'consumer1' or 'producer10'.
        Returns:
          S_OK: with the value of the messenger Id or S_ERROR if the messenger was not added,
                cause the same id already exists.
        """
        self.lock.acquire()
        try:
            conn = getMQService(mqURI)
            dest = getDestinationAddress(mqURI)
            connector = self.__getConnector(conn)

            if not self.__removeMessenger(conn, dest, messengerId):
                return S_ERROR(EMQCONN, "Failed to stop the connection!The messenger %s does not exist!" % messengerId)
            else:
                if "consumer" in messengerId:
                    result = self.unsubscribe(connector, destination=dest, messengerId=messengerId)
                    if not result["OK"]:
                        return result
            if not self.__connectionExists(conn):
                return self.disconnect(connector)
            return S_OK()
        finally:
            self.lock.release()

    def getAllMessengers(self):
        """Function returns a list of all messengers registered in connection storage.

        Returns:
          S_OK or S_ERROR: with the list of strings in the pseudo-path format e.g.
                ['blabla.cern.ch/queue/test1/consumer1','blabal.cern.ch/topic/test2/producer2']
        """
        self.lock.acquire()
        try:
            return S_OK(self.__getAllMessengersInfo())
        finally:
            self.lock.release()

    def removeAllConnections(self):
        """Function removes all existing connections and calls the disconnect
            for connectors.

        Returns:
          S_OK or S_ERROR:
        """
        self.lock.acquire()
        try:
            connections = self.__getAllConnections()
            for conn in connections:
                connector = self.__getConnector(conn)
                if connector:
                    self.disconnect(connector)
            self.__connectionStorage = {}
            return S_OK()
        finally:
            self.lock.release()

    # Set of 'private' helper functions to access and modify the message queue connection storage.
    def __getConnection(self, mqConnection):
        """Function returns message queue connection from the storage.
        Args:
          mqConnection(str): message queue connection name.
        Returns:
          dict:
        """
        return self.__connectionStorage.get(mqConnection, {})

    def __getAllConnections(self):
        """Function returns a list of all connection names in the storage
        Returns:
          list:
        """
        return list(self.__connectionStorage)

    def __getConnector(self, mqConnection):
        """Function returns MQConnector from the storage.
        Args:
          mqConnection(str): message queue connection name.
        Returns:
          MQConnector or None
        """
        return self.__getConnection(mqConnection).get("MQConnector", None)

    def __setConnector(self, mqConnection, connector):
        """Function returns MQConnector from the storage.
        Args:
          mqConnection(str): message queue connection name.
          connector(MQConnector):
        Returns:
          bool: False if connection does not exit
        """
        connDict = self.__getConnection(mqConnection)
        if not connDict:
            return False
        connDict["MQConnector"] = connector
        return True

    def __getDestinations(self, mqConnection):
        """Function returns dict with destinations (topics and queues) for given connection.
        Args:
          mqConnection(str): message queue connection name.
        Returns:
          dict: of form {'/queue/queue1':['producer1','consumer2']} or {}
        """
        return self.__getConnection(mqConnection).get("destinations", {})

    def __getMessengersId(self, mqConnection, mqDestination):
        """Function returns list of messengers for given connection and given destination.
        Args:
          mqConnection(str): message queue connection name.
          mqDestination(str): message queue or topic name e.g. '/queue/myQueue1' .
        Returns:
          list: of form ['producer1','consumer2'] or []
        """
        return self.__getDestinations(mqConnection).get(mqDestination, [])

    def __getMessengersIdWithType(self, mqConnection, mqDestination, messengerType):
        """Function returns list of messnager for given connection, destination and type.
        Args:
          mqConnection(str): message queue connection name.
          mqDestination(str): message queue or topic name e.g. '/queue/myQueue1' .
          messengerType(str): 'consumer' or 'producer'
        Returns:
          list: of form ['producer1','producer2'], ['consumer8', 'consumer20'] or []
        """
        return [p for p in self.__getMessengersId(mqConnection, mqDestination) if messengerType in p]

    def __getAllMessengersId(self):
        """Function returns list of all messengers ids.
            The list can contain duplicates because the same
            producer id can be used for different queues.
        Args:
        Returns:
          list: of form ['producer1','consumer1', 'producer1'] or []
        """
        return [
            m
            for c in self.__connectionStorage.keys()
            for d in self.__getDestinations(c)
            for m in self.__getMessengersId(c, d)
        ]

    def __getAllMessengersIdWithType(self, messengerType):
        """Function returns list of all messengers ids for given messengerType
        Args:
          messengerType(str): 'consumer' or 'producer'
        Returns:
          list: of form ['producer1','producer2'], ['consumer8', 'consumer20'] or []
        """
        return [p for p in self.__getAllMessengersId() if messengerType in p]

    def __getAllMessengersInfo(self):
        """Function returns list of all messengers in the pseudo-path format.
        Returns:
          list: of form ['blabla.cern.ch/queue/myQueue1/producer1','bibi.in2p3.fr/topic/myTopic331/consumer3'] or []
        """

        def output(connection, dest, messenger):
            return str(connection) + str(dest) + "/" + str(messenger)

        return [
            output(c, d, m)
            for c in self.__connectionStorage.keys()
            for d in self.__getDestinations(c)
            for m in self.__getMessengersId(c, d)
        ]

    def __connectionExists(self, mqConnection):
        """Function checks if given connection exists in the connection storage.
        Args:
          mqConnection(str): message queue connection name.
        Returns:
          bool:
        """
        return mqConnection in self.__connectionStorage

    def __destinationExists(self, mqConnection, mqDestination):
        """Function checks if given destination(queue or topic) exists in the connection storage.
        Args:
          mqConnection(str): message queue connection name.
          mqDestination(str): message queue or topic name e.g. '/queue/myQueue1' .
        Returns:
          bool:
        """
        return mqDestination in self.__getDestinations(mqConnection)

    def __messengerExists(self, mqConnection, mqDestination, messengerId):
        """Function checks if given messenger(producer or consumer) exists in the connection storage.
        Args:
          mqConnection(str): message queue connection name.
          mqDestination(str): message queue or topic name e.g. '/queue/myQueue1' .
          messengerId(str): messenger name e.g. 'consumer1', 'producer4' .
        Returns:
          bool:
        """
        return messengerId in self.__getMessengersId(mqConnection, mqDestination)

    def __addMessenger(self, mqConnection, destination, messengerId):
        """Function adds a messenger(producer or consumer) to given connection and destination.
            If connection or/and destination do not exist, they are created as well.
        Args:
          mqConnection(str): message queue connection name.
          mqDestination(str): message queue or topic name e.g. '/queue/myQueue1' .
          messengerId(str): messenger name e.g. 'consumer1', 'producer4'.
        Returns:
          bool: True if messenger is added or False if the messenger already exists.
        """
        if self.__messengerExists(mqConnection, destination, messengerId):
            return False
        if self.__connectionExists(mqConnection):
            if self.__destinationExists(mqConnection, destination):
                self.__getMessengersId(mqConnection, destination).append(messengerId)
            else:
                self.__getDestinations(mqConnection)[destination] = [messengerId]
        else:
            self.__connectionStorage[mqConnection] = {"MQConnector": None, "destinations": {destination: [messengerId]}}
        return True

    def __removeMessenger(self, mqConnection, destination, messengerId):
        """Function removes  messenger(producer or consumer) from given connection and destination.
            If it is the last messenger in given destination and/or connection they are removed as well..
        Args:
          mqConnection(str): message queue connection name.
          mqDestination(str): message queue or topic name e.g. '/queue/myQueue1' .
          messengerId(str): messenger name e.g. 'consumer1', 'producer4'.
        Returns:
          bool: True if messenger is removed or False if the messenger was not in the storage.
        """
        messengers = self.__getMessengersId(mqConnection, destination)
        destinations = self.__getDestinations(mqConnection)
        if messengerId in messengers:
            messengers.remove(messengerId)
            if not messengers:  # If no more messengers we remove the destination.
                destinations.pop(destination)
                if not destinations:  # If no more destinations we remove the connection
                    self.__connectionStorage.pop(mqConnection)
            return True
        else:
            return False  # messenger was not in the storage



#########################################################################################
# LSF.py
# 10.11.2014
# Author: A.T.
#########################################################################################

""" LSF.py is a DIRAC independent class representing LSF batch system.
    LSF objects are used as backend batch system representation for
    LocalComputingElement and SSHComputingElement classes
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
import re
import subprocess
import shlex
import os

__RCSID__ = "$Id$"


class LSF(object):
    def submitJob(self, **kwargs):
        """Submit nJobs to the condor batch system"""

        resultDict = {}

        MANDATORY_PARAMETERS = ["Executable", "OutputDir", "ErrorDir", "WorkDir", "SubmitOptions", "Queue"]
        for argument in MANDATORY_PARAMETERS:
            if argument not in kwargs:
                resultDict["Status"] = -1
                resultDict["Message"] = "No %s" % argument
                return resultDict

        nJobs = kwargs.get("NJobs", 1)
        preamble = kwargs.get("Preamble")

        outputs = []
        outputDir = kwargs["OutputDir"]
        errorDir = kwargs["ErrorDir"]
        executable = kwargs["Executable"]
        queue = kwargs["Queue"]
        submitOptions = kwargs["SubmitOptions"]
        outputDir = os.path.expandvars(outputDir)
        errorDir = os.path.expandvars(errorDir)
        executable = os.path.expandvars(executable)
        for _i in range(int(nJobs)):
            cmd = "%s; " % preamble if preamble else ""
            cmd += "bsub -o %s -e %s -q %s -J DIRACPilot %s %s" % (
                outputDir,
                errorDir,
                queue,
                submitOptions,
                executable,
            )
            sp = subprocess.Popen(
                cmd,
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                universal_newlines=True,
            )
            output, error = sp.communicate()
            status = sp.returncode
            if status == 0:
                outputs.append(output)
            else:
                break

        if outputs:
            resultDict["Status"] = 0
            resultDict["Jobs"] = []
            for output in outputs:
                match = re.search(r"Job <(\d*)>", output)
                if match:
                    resultDict["Jobs"].append(match.groups()[0])
        else:
            resultDict["Status"] = status
            resultDict["Message"] = error

        return resultDict

    def killJob(self, **kwargs):
        """Kill jobs in the given list"""

        resultDict = {}

        MANDATORY_PARAMETERS = ["JobIDList"]
        for argument in MANDATORY_PARAMETERS:
            if argument not in kwargs:
                resultDict["Status"] = -1
                resultDict["Message"] = "No %s" % argument
                return resultDict

        jobIDList = kwargs.get("JobIDList")
        if not jobIDList:
            resultDict["Status"] = -1
            resultDict["Message"] = "Empty job list"
            return resultDict

        successful = []
        failed = []
        errors = ""
        for job in jobIDList:
            sp = subprocess.Popen(
                shlex.split("bkill %s" % job),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                universal_newlines=True,
            )
            output, error = sp.communicate()
            status = sp.returncode
            if status != 0:
                failed.append(job)
                errors += error
            else:
                successful.append(job)

        resultDict["Status"] = 0
        if failed:
            resultDict["Status"] = 1
            resultDict["Message"] = error
        resultDict["Successful"] = successful
        resultDict["Failed"] = failed
        return resultDict

    def getCEStatus(self, **kwargs):
        """Method to return information on running and pending jobs."""

        resultDict = {}

        MANDATORY_PARAMETERS = ["Queue"]
        for argument in MANDATORY_PARAMETERS:
            if argument not in kwargs:
                resultDict["Status"] = -1
                resultDict["Message"] = "No %s" % argument
                return resultDict

        queue = kwargs["Queue"]

        cmd = "bjobs -q %s -a" % queue
        sp = subprocess.Popen(
            shlex.split(cmd),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )
        output, error = sp.communicate()
        status = sp.returncode

        if status != 0:
            resultDict["Status"] = status
            resultDict["Message"] = error
            return resultDict

        waitingJobs = 0
        runningJobs = 0
        lines = output.split("\n")
        for line in lines:
            if line.count("PEND") or line.count("PSUSP"):
                waitingJobs += 1
            if line.count("RUN") or line.count("USUSP"):
                runningJobs += 1

        # Final output
        resultDict["Status"] = 0
        resultDict["Waiting"] = waitingJobs
        resultDict["Running"] = runningJobs
        return resultDict

    def getJobStatus(self, **kwargs):
        """Get the status information for the given list of jobs"""

        resultDict = {}

        MANDATORY_PARAMETERS = ["JobIDList"]
        for argument in MANDATORY_PARAMETERS:
            if argument not in kwargs:
                resultDict["Status"] = -1
                resultDict["Message"] = "No %s" % argument
                return resultDict

        jobIDList = kwargs["JobIDList"]
        if not jobIDList:
            resultDict["Status"] = -1
            resultDict["Message"] = "Empty job list"
            return resultDict

        cmd = "bjobs " + " ".join(jobIDList)
        sp = subprocess.Popen(
            shlex.split(cmd),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )
        output, error = sp.communicate()
        status = sp.returncode

        if status != 0:
            resultDict["Status"] = status
            resultDict["Message"] = error
            return resultDict

        output = output.replace("\r", "")
        lines = output.split("\n")
        statusDict = {}
        for job in jobIDList:
            statusDict[job] = "Unknown"
            for line in lines:
                if line.find(job) != -1:
                    if line.find("UNKWN") != -1:
                        statusDict[job] = "Unknown"
                    else:
                        lsfStatus = line.split()[2]
                        if lsfStatus in ["DONE", "EXIT"]:
                            statusDict[job] = "Done"
                        elif lsfStatus in ["RUN", "SSUSP"]:
                            statusDict[job] = "Running"
                        elif lsfStatus in ["PEND", "PSUSP"]:
                            statusDict[job] = "Waiting"

        # Final output
        status = 0
        resultDict["Status"] = 0
        resultDict["Jobs"] = statusDict
        return resultDict



""" DIRAC Notification Client class encapsulates the methods exposed
    by the Notification service.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six
from DIRAC import gLogger, S_ERROR
from DIRAC.Core.Base.Client import Client, createClient
from DIRAC.Core.Utilities.Mail import Mail


@createClient("Framework/Notification")
class NotificationClient(Client):
    def __init__(self, **kwargs):
        """Notification Client constructor"""
        super(NotificationClient, self).__init__(**kwargs)

        self.log = gLogger.getSubLogger("NotificationClient")
        self.setServer("Framework/Notification")

    def sendMail(self, addresses, subject, body, fromAddress=None, localAttempt=True, html=False):
        """Send an e-mail with subject and body to the specified address. Try to send
        from local area before central service by default.
        """
        self.log.verbose(
            "Received signal to send the following mail to %s:\nSubject = %s\n%s" % (addresses, subject, body)
        )
        result = S_ERROR()

        addresses = [addresses] if isinstance(addresses, six.string_types) else list(addresses)
        for address in addresses:

            if localAttempt:
                try:
                    m = Mail()
                    m._subject = subject
                    m._message = body
                    m._mailAddress = address
                    m._html = html
                    if fromAddress:
                        m._fromAddress = fromAddress
                    result = m._send()
                except Exception as x:
                    self.log.warn("Sending mail failed with exception:\n%s" % (str(x)))

                if result["OK"]:
                    self.log.verbose(
                        "Mail sent successfully from local host to %s with subject %s" % (address, subject)
                    )
                    self.log.debug(result["Value"])
                    return result

                self.log.warn(
                    "Could not send mail with the following message:\n%s\n will attempt to send via NotificationService"
                    % result["Message"]
                )

            result = self._getRPC().sendMail(address, subject, body, str(fromAddress))
            if not result["OK"]:
                self.log.error("Could not send mail via central Notification service", result["Message"])
                return result
            else:
                self.log.verbose(result["Value"])

        return result

    def sendSMS(self, userName, body, fromAddress=None):
        """Send an SMS with body to the specified DIRAC user name."""
        self.log.verbose("Received signal to send the following SMS to %s:\n%s" % (userName, body))
        result = self._getRPC().sendSMS(userName, body, str(fromAddress))
        if not result["OK"]:
            self.log.error("Could not send SMS via central Notification service", result["Message"])
        else:
            self.log.verbose(result["Value"])

        return result

    ###########################################################################
    # ALARMS
    ###########################################################################

    def newAlarm(self, subject, status, notifications, assignee, body, priority, alarmKey=""):
        if not isinstance(notifications, (list, tuple)):
            return S_ERROR(
                "Notifications parameter has to be a list or a tuple with a combination of [ 'Web', 'Mail', 'SMS' ]"
            )
        alarmDef = {
            "subject": subject,
            "status": status,
            "notifications": notifications,
            "assignee": assignee,
            "priority": priority,
            "body": body,
        }
        if alarmKey:
            alarmDef["alarmKey"] = alarmKey
        return self._getRPC().newAlarm(alarmDef)

    def updateAlarm(self, id=-1, alarmKey="", comment=False, modDict={}):
        if id == -1 and not alarmKey:
            return S_ERROR("Need either alarm id or key to update an alarm!")
        updateReq = {"comment": comment, "modifications": modDict}
        if id != -1:
            updateReq["id"] = id
        if alarmKey:
            updateReq["alarmKey"] = alarmKey
        return self._getRPC().updateAlarm(updateReq)

    ###########################################################################
    # MANAGE NOTIFICATIONS
    ###########################################################################

    def addNotificationForUser(self, user, message, lifetime=604800, deferToMail=True):
        try:
            lifetime = int(lifetime)
        except Exception:
            return S_ERROR("Message lifetime has to be a non decimal number")
        return self._getRPC().addNotificationForUser(user, message, lifetime, deferToMail)



""" DIRAC Production DB

    Production database is used to collect and serve the necessary information
    in order to automate the task of transformation preparation for high level productions.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# # imports
import six
import json
import threading

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.Base.DB import DB
from DIRAC.ProductionSystem.Utilities.ProdValidator import ProdValidator
from DIRAC.ProductionSystem.Utilities.ProdTransManager import ProdTransManager
from DIRAC.Core.Utilities.List import intListToString


__RCSID__ = "$Id$"

MAX_ERROR_COUNT = 10

#############################################################################


class ProductionDB(DB):
    """ProductionDB class"""

    def __init__(self, dbname=None, dbconfig=None, dbIn=None):
        """The standard constructor takes the database name (dbname) and the name of the
        configuration section (dbconfig)
        """
        if not dbname:
            dbname = "ProductionDB"
        if not dbconfig:
            dbconfig = "Production/ProductionDB"

        if not dbIn:
            DB.__init__(self, dbname, dbconfig)

        self.lock = threading.Lock()

        self.prodValidator = ProdValidator()
        self.ProdTransManager = ProdTransManager()

        self.PRODPARAMS = [
            "ProductionID",
            "ProductionName",
            "Description",
            "CreationDate",
            "LastUpdate",
            "AuthorDN",
            "AuthorGroup",
            "Status",
        ]

        self.TRANSPARAMS = ["TransformationID", "ProductionID", "LastUpdate", "InsertedTime"]

        self.TRANSLINKSPARAMS = ["TransformationID", "ParentTransformationID", "ProductionID"]

        self.PRODSTEPSPARAMS = [
            "StepID",
            "Name",
            "Description",
            "LongDescription",
            "Body",
            "Type",
            "Plugin",
            "AgentType",
            "GroupSize",
            "InputQuery",
            "OutputQuery",
            "LastUpdate",
            "InsertedTime",
        ]

        self.statusActionDict = {
            "New": None,
            "Active": "startTransformation",
            "Stopped": "stopTransformation",
            "Cleaned": "cleanTransformation",
        }

    def addProduction(self, prodName, prodDescription, authorDN, authorGroup, connection=False):
        """Create new production starting from its description

        :param str prodName: a string with the Production name
        :param str prodDescription: a json object with the Production description
        :param str authorDN: string with the author DN
        :param str authorGroup: string with author group
        """
        connection = self.__getConnection(connection)
        res = self._getProductionID(prodName, connection=connection)
        if res["OK"]:
            return S_ERROR("Production with name %s already exists with ProductionID = %d" % (prodName, res["Value"]))
        elif res["Message"] != "Production does not exist":
            return res

        self.lock.acquire()

        req = (
            "INSERT INTO Productions (ProductionName,Description,CreationDate,LastUpdate,\
                                    AuthorDN,AuthorGroup,Status)\
                                VALUES ('%s','%s',UTC_TIMESTAMP(),UTC_TIMESTAMP(),'%s','%s','New');"
            % (prodName, prodDescription, authorDN, authorGroup)
        )

        res = self._update(req, connection)
        if not res["OK"]:
            self.lock.release()
            return res
        prodID = int(res["lastRowId"])
        self.lock.release()

        return S_OK(prodID)

    def getProductions(
        self,
        condDict=None,
        older=None,
        newer=None,
        timeStamp="LastUpdate",
        orderAttribute=None,
        limit=None,
        offset=None,
        connection=False,
    ):
        """Get parameters of all the Productions with support for the web standard structure

        :param dict condDict:
        :param str older:
        :param str newer:
        :param str timeStamp:
        :param str orderAttribute:
        :param int limit:
        :param int offset:
        :param bool connection:
        :return:
        """

        connection = self.__getConnection(connection)
        req = "SELECT %s FROM Productions %s" % (
            intListToString(self.PRODPARAMS),
            self.buildCondition(condDict, older, newer, timeStamp, orderAttribute, limit, offset=offset),
        )
        res = self._query(req, connection)
        if not res["OK"]:
            return res

        webList = []
        resultList = []
        for row in res["Value"]:
            # Prepare the structure for the web
            rList = [str(item) if not isinstance(item, six.integer_types) else item for item in row]
            prodDict = dict(zip(self.PRODPARAMS, row))
            webList.append(rList)
            resultList.append(prodDict)
        result = S_OK(resultList)
        result["Records"] = webList
        result["ParameterNames"] = self.PRODPARAMS
        return result

    def getProduction(self, prodName, connection=False):
        """Get the Production definition

        :param str prodName: the Production name or ID
        """
        res = self._getConnectionProdID(connection, prodName)
        if not res["OK"]:
            return res
        connection = res["Value"]["Connection"]
        prodID = res["Value"]["ProductionID"]
        res = self.getProductions(condDict={"ProductionID": prodID}, connection=connection)
        if not res["OK"]:
            return res
        if not res["Value"]:
            return S_ERROR("Production %s did not exist" % prodName)
        return S_OK(res["Value"][0])

    def getProductionParameters(self, prodName, parameters, connection=False):
        """Get the requested parameters for a supplied production

        :param str prodName: the Production name or ID
        :param str parameters: any valid production parameter in self.PRODPARAMS
        """
        if isinstance(parameters, six.string_types):
            parameters = [parameters]
        res = self.getProduction(prodName, connection=connection)
        if not res["OK"]:
            return res
        prodParams = res["Value"]
        paramDict = {}
        for reqParam in parameters:
            if reqParam not in prodParams:
                return S_ERROR("Parameter %s not defined for production" % reqParam)
            paramDict[reqParam] = prodParams[reqParam]
        if len(paramDict) == 1:
            return S_OK(paramDict[reqParam])
        return S_OK(paramDict)

    ###########################################################################
    #
    # These methods manipulate the ProductionSteps table
    #

    def getProductionStep(self, stepID, connection=False):
        """It returns the ProductionStep corresponding to the stepID

        :param int stepID: the ID of the production Step stored in the ProductionSteps table
        :return: the attributes of Production Step corresponding to the stepID
        """
        connection = self.__getConnection(connection)
        req = "SELECT %s FROM ProductionSteps WHERE StepID = %s" % (intListToString(self.PRODSTEPSPARAMS), str(stepID))
        res = self._query(req, connection)
        if not res["OK"]:
            return res
        if not res["Value"]:
            return S_ERROR("ProductionStep %s did not exist" % str(stepID))
        return S_OK(res["Value"][0])

    def addProductionStep(
        self,
        stepName,
        stepDescription,
        stepLongDescription,
        stepBody,
        stepType,
        stepPlugin,
        stepAgentType,
        stepGroupSize,
        stepInputquery,
        stepOutputquery,
        connection=False,
    ):
        """Add a Production Step

        :param str stepName: name of the production Step
        :param str stepDescription: description of the production Step
        :param str stepLongDescription: long description of the production Step
        :param str stepBody: body of the production Step
        :param str stepType: type of the production Step
        :param str stepPlugin: plugin to be used for the production Step
        :param str stepAgentType: agent type to be used for the production Step
        :param int stepGroupSize: group size of the production Step
        :param str stepInputquery: InputQuery of the production Step
        :param str stepOutputquery: OutputQuery of the production Step
        :return:
        """
        connection = self.__getConnection(connection)
        self.lock.acquire()
        req = "INSERT INTO ProductionSteps (Name,Description,LongDescription,Body,Type,Plugin,AgentType,GroupSize,\
                                    InputQuery,OutputQuery,LastUpdate,InsertedTime)\
                                VALUES ('%s','%s', '%s', %s, '%s', '%s', '%s', '%s', '%s', '%s',\
                                UTC_TIMESTAMP(),UTC_TIMESTAMP());" % (
            stepName,
            stepDescription,
            stepLongDescription,
            stepBody,
            stepType,
            stepPlugin,
            stepAgentType,
            stepGroupSize,
            stepInputquery,
            stepOutputquery,
        )

        res = self._update(req, connection)
        if not res["OK"]:
            self.lock.release()
            return res
        stepID = int(res["lastRowId"])
        self.lock.release()

        return S_OK(stepID)

        ###########################################################################

    #
    # These methods manipulate the ProductionTransformations table
    #

    def getProductionTransformations(
        self,
        prodName,
        condDict=None,
        older=None,
        newer=None,
        timeStamp="CreationTime",
        orderAttribute=None,
        limit=None,
        offset=None,
        connection=False,
    ):
        """Gets the transformations of a given Production

        :param str prodName: the Production name or ID
        :return:
        """
        res = self._getConnectionProdID(connection, prodName)
        if not res["OK"]:
            return res
        connection = res["Value"]["Connection"]
        prodID = res["Value"]["ProductionID"]
        condDict = {"ProductionID": prodID}

        req = "SELECT %s FROM ProductionTransformations %s" % (
            intListToString(self.TRANSPARAMS),
            self.buildCondition(condDict, older, newer, timeStamp, orderAttribute, limit, offset=offset),
        )
        res = self._query(req, connection)
        if not res["OK"]:
            return res

        webList = []
        resultList = []
        for row in res["Value"]:
            # Prepare the structure for the web
            rList = [str(item) if not isinstance(item, six.integer_types) else item for item in row]
            transDict = dict(zip(self.TRANSPARAMS, row))
            webList.append(rList)
            resultList.append(transDict)
        result = S_OK(resultList)
        result["Records"] = webList
        result["ParameterNames"] = self.TRANSPARAMS
        return result

    def __setProductionStatus(self, prodID, status, connection=False):
        """Set the Status to the Production

        :param int prodID: the ProductionID
        :param str status: the Production status
        """
        req = "UPDATE Productions SET Status='%s', LastUpdate=UTC_TIMESTAMP() WHERE ProductionID=%d" % (status, prodID)
        return self._update(req, connection)

    # This is to be replaced by startProduction, stopProduction etc.
    def setProductionStatus(self, prodName, status, connection=False):
        """Set the status to the production and to all the associated transformations

        :param str prodName: the Production name or ID
        :param str status: the Production status
        """
        res = self._getConnectionProdID(connection, prodName)
        gLogger.error(res)
        if not res["OK"]:
            return res
        connection = res["Value"]["Connection"]
        prodID = res["Value"]["ProductionID"]

        res = self.__setProductionStatus(prodID, status, connection=connection)
        if not res["OK"]:
            return res

        res = self.ProdTransManager.executeActionOnTransformations(prodID, self.statusActionDict[status])
        if not res["OK"]:
            gLogger.error(res["Message"])

        return S_OK()

    def startProduction(self, prodName, connection=False):
        """Instantiate and start the transformations belonging to the production

        :param str prodName: the Production name or ID
        """

        res = self._getConnectionProdID(connection, prodName)
        gLogger.error(res)
        if not res["OK"]:
            return res
        connection = res["Value"]["Connection"]
        prodID = res["Value"]["ProductionID"]

        # Instantiate the transformations according to the description
        res = self.getProductionParameters(prodName, "Description")
        if not res["OK"]:
            return res

        prodDescription = json.loads(res["Value"])
        transIDs = []
        for step in sorted(prodDescription):
            res = self.ProdTransManager.addTransformationStep(prodDescription[step]["stepID"], prodID)
            # If the addition of at least one step failes, clean the already added transformation steps from the TS
            if not res["OK"]:
                self.ProdTransManager.deleteTransformations(transIDs)
                return S_ERROR(res["Message"])
            transID = res["Value"]
            prodDescription[step]["transID"] = transID
            transIDs.append(transID)

        for step in prodDescription:
            transID = prodDescription[step]["transID"]
            parentTransIDs = []
            if "parentStep" in prodDescription[step]:
                for parentStep in prodDescription[step]["parentStep"]:
                    parentTransID = prodDescription[parentStep]["transID"]
                    parentTransIDs.append(parentTransID)

            res = self.addTransformationsToProduction(prodID, transID, parentTransIDs=parentTransIDs)
            # If adding the transformations to production fails, clean all the transformation steps from the TS
            if not res["OK"]:
                self.ProdTransManager.deleteTransformations(transIDs)
                # Clean the production
                self.deleteProduction(prodID)
                return S_ERROR(res["Message"])

        res = self.__setProductionStatus(prodID, "Active", connection=connection)
        if not res["OK"]:
            return res

        res = self.ProdTransManager.executeActionOnTransformations(prodID, self.statusActionDict["Active"])
        if not res["OK"]:
            gLogger.error(res["Message"])

        return S_OK()

    def __deleteProduction(self, prodID, connection=False):
        """Delete a given production from the Productions table

        :param int prodID: ProductionID
        """
        req = "DELETE FROM Productions WHERE ProductionID=%d;" % prodID
        return self._update(req, connection)

    def __deleteProductionTransformations(self, prodID, connection=False):
        """Remove all the transformations of the specified production from the TS and from the PS

        :param int prodID: the ProductionID
        """

        # Remove transformations from the TS
        gLogger.notice("Deleting transformations of Production %s from the TS" % prodID)
        res = self.ProdTransManager.deleteProductionTransformations(prodID)
        if not res["OK"]:
            gLogger.error("Failed to delete production transformations from the TS", res["Message"])

        # Remove transformations from the PS
        req = "DELETE FROM ProductionTransformationLinks WHERE ProductionID = %d;" % prodID
        res = self._update(req, connection)
        if not res["OK"]:
            gLogger.error("Failed to delete production transformation links from the PS", res["Message"])

        req = "DELETE FROM ProductionTransformations WHERE ProductionID = %d;" % prodID
        res = self._update(req, connection)
        if not res["OK"]:
            gLogger.error("Failed to delete production transformations from the PS", res["Message"])

        return res

    def cleanProduction(self, prodName, author="", connection=False):
        """Clean the production specified by name or id

        :param str prodName: the Production name or ID
        """
        res = self._getConnectionProdID(connection, prodName)
        if not res["OK"]:
            return res
        connection = res["Value"]["Connection"]
        prodID = res["Value"]["ProductionID"]
        res = self.__deleteProductionTransformations(prodID, connection=connection)
        if not res["OK"]:
            return res

        return S_OK(prodID)

    def deleteProduction(self, prodName, author="", connection=False):
        """Remove the production specified by name or id

        :param str prodName: the Production name or ID
        """
        res = self._getConnectionProdID(connection, prodName)
        gLogger.error(res)
        if not res["OK"]:
            return res
        connection = res["Value"]["Connection"]
        prodID = res["Value"]["ProductionID"]
        res = self.cleanProduction(prodID, author=author, connection=connection)
        if not res["OK"]:
            return res
        res = self.__deleteProduction(prodID, connection=connection)
        if not res["OK"]:
            return res

        return S_OK()

    def addTransformationsToProduction(self, prodName, transIDs, parentTransIDs=None, connection=False):
        """Check the production validity and the add the transformations to the production

        :param str prodName: the Production name or ID
        :param list transIDs: the list of transformations to be added to the production
        """
        gLogger.info(
            "ProductionDB.addTransformationsToProduction: \
         Attempting to add %s transformations to production: %s"
            % (transIDs, prodName)
        )

        if not transIDs:
            return S_ERROR("Zero length transformation list")
        res = self._getConnectionProdID(connection, prodName)
        if not res["OK"]:
            return res
        connection = res["Value"]["Connection"]
        prodID = res["Value"]["ProductionID"]

        # Check the production validity
        gLogger.notice("Checking if production is valid")

        # Set transformations in list format
        if parentTransIDs:
            if not isinstance(parentTransIDs, list):
                parentTransIDs = [parentTransIDs]

        if not isinstance(transIDs, list):
            transIDs = [transIDs]

        # Check the status of the transformations (must be 'New')
        for transID in transIDs:
            res = self.prodValidator.checkTransStatus(transID)
            if not res["OK"]:
                gLogger.error("Production is not valid:", res["Message"])
                return res

        # Check if transformations are linked to their parent transformations, if they have one(s)
        if parentTransIDs:
            for transID in transIDs:
                for parentTransID in parentTransIDs:
                    gLogger.notice(
                        "Checking if transformation %s is linked to the parent transformation %s"
                        % (transID, parentTransID)
                    )
                    res = self.prodValidator.checkTransDependency(transID, parentTransID)
                    if not res["OK"]:
                        gLogger.error("Production is not valid:", res["Message"])
                        return res

        gLogger.notice("Production %s is valid" % prodName)

        res = self.__addTransformations(prodID, transIDs, connection=connection)
        if not res["OK"]:
            msg = "Failed to add transformations %s to production %s: %s" % (transIDs, prodID, res["Message"])
            return S_ERROR(msg)
        # Add the transformation links (tranformation to parent transformation) for the given production
        res = self.__addTransformationLinks(prodID, transIDs, parentTransIDs=parentTransIDs, connection=connection)
        if not res["OK"]:
            msg = "Failed to add production transformations links to transformations %s: %s" % (
                transIDs,
                res["Message"],
            )
            return S_ERROR(msg)

        # Update the status of the transformation to be in sync with the status of the production
        res = self.getProduction(prodID)
        prodStatus = res["Value"]["Status"]

        # Execute action on transformations according to the production status
        res = self.ProdTransManager.executeActionOnTransformations(prodID, self.statusActionDict[prodStatus])
        if not res["OK"]:
            gLogger.error(res["Message"])

        return S_OK()

    def __addTransformationLinks(self, prodID, transIDs, parentTransIDs=None, connection=False):
        """Insert the transformations in the ProductionTransformationLinks table

        :param prodName: the Production name or ID
        :param transIDs: the list of transformations to be associated ProductionTransformationLinks table
        """
        req = "INSERT INTO ProductionTransformationLinks \
           (TransformationID,ParentTransformationID,ProductionID) VALUES"

        # Insert transformations and the corresponding parent transformations
        if parentTransIDs:
            for transID in transIDs:
                for parentTransID in parentTransIDs:
                    req = "%s (%d,%d,%d)," % (req, transID, parentTransID, prodID)
        # If parent transformations are not defined, just insert transformations and use the parent transformation default
        # value -1
        else:
            req = "INSERT INTO ProductionTransformationLinks (TransformationID,ProductionID) VALUES"
            for transID in transIDs:
                req = "%s (%d,%d)," % (req, transID, prodID)

        gLogger.notice(req)
        req = req.rstrip(",")
        res = self._update(req, connection)
        if not res["OK"]:
            return res

        return S_OK()

    def __addTransformations(self, prodID, transIDs, connection=False):
        """Insert the transformations in the ProductionTransformations table

        :param str prodName: the Production name or ID
        :param list transIDs: the list of transformations to be added to the production
        """
        req = "INSERT INTO ProductionTransformations \
           (ProductionID,TransformationID,LastUpdate,InsertedTime) VALUES"
        for transID in transIDs:
            req = "%s (%d,%d,UTC_TIMESTAMP(),UTC_TIMESTAMP())," % (req, prodID, transID)
            gLogger.notice(req)
        req = req.rstrip(",")
        res = self._update(req, connection)
        if not res["OK"]:
            return res

        return S_OK()

    def _getProductionID(self, prodName, connection=False):
        """Method returns ID of production specified by the prodName

        :param str prodName: the Production name
        """
        try:
            prodName = int(prodName)
            cmd = "SELECT ProductionID from Productions WHERE ProductionID=%d;" % prodName
        except Exception:
            if not isinstance(prodName, six.string_types):
                return S_ERROR("Production should be ID or name")
            cmd = "SELECT ProductionID from Productions WHERE ProductionName='%s';" % prodName
        res = self._query(cmd, connection)
        if not res["OK"]:
            gLogger.error("Failed to obtain production ID for production", "%s: %s" % (prodName, res["Message"]))
            return res
        elif not res["Value"]:
            gLogger.verbose("Production %s does not exist" % (prodName))
            return S_ERROR("Production does not exist")
        return S_OK(res["Value"][0][0])

    def __getConnection(self, connection):
        """Get the MySQL connection

        :param bool connection: the DB connection
        """
        if connection:
            return connection
        res = self._getConnection()
        if res["OK"]:
            return res["Value"]
        gLogger.warn("Failed to get MySQL connection", res["Message"])
        return connection

    def _getConnectionProdID(self, connection, prodName):
        """Get the Production ID for a given production

        :param bool connection: the DB connection
        :param prodName: the Production name
        """
        connection = self.__getConnection(connection)
        res = self._getProductionID(prodName, connection=connection)
        if not res["OK"]:
            gLogger.error("Failed to get ID for production %s: %s" % (prodName, res["Message"]))
            return res
        prodID = res["Value"]
        resDict = {"Connection": connection, "ProductionID": prodID}
        return S_OK(resDict)
